
from pathlib import Path

from alnitak import prog as Prog
from alnitak import config
from alnitak.tests import setup


def test_config_default():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.ok

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))

    t_a1 = setup.create_tlsa_obj('311', '12725', 'tcp', 'a.com')
    t_a2 = setup.create_tlsa_obj('201', '12725', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [], [t_a1, t_a2])

    t_b1 = setup.create_tlsa_obj('311', '12780', 'udp', 'b.com')
    t_b2 = setup.create_tlsa_obj('201', '12780', 'sctp', 'A.b.com')
    tb = setup.create_target_obj('b.com', api, [], [t_b1, t_b2])

    t_c1 = setup.create_tlsa_obj('311', '12722', 'tcp', 'A.c.com')
    t_c2 = setup.create_tlsa_obj('311', '12723', 'tcp', 'B.c.com')
    tc = setup.create_target_obj('c.com', api, [], [t_c1, t_c2])

    assert prog.target_list == [ta, tb, tc]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le


def test_config1():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.config1)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.ok

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))

    t_a200 = setup.create_tlsa_obj('200', '12725', 'tcp', 'a.com')
    t_a201 = setup.create_tlsa_obj('201', '12725', 'tcp', 'a.com')
    t_a202 = setup.create_tlsa_obj('202', '12725', 'tcp', 'a.com')
    t_a210 = setup.create_tlsa_obj('210', '12725', 'tcp', 'a.com')
    t_a211 = setup.create_tlsa_obj('211', '12725', 'tcp', 'a.com')
    t_a212 = setup.create_tlsa_obj('212', '12725', 'tcp', 'a.com')
    t_a300 = setup.create_tlsa_obj('300', '12725', 'tcp', 'a.com')
    t_a301 = setup.create_tlsa_obj('301', '12725', 'tcp', 'a.com')
    t_a302 = setup.create_tlsa_obj('302', '12725', 'tcp', 'a.com')
    t_a310 = setup.create_tlsa_obj('310', '12725', 'tcp', 'a.com')
    t_a311 = setup.create_tlsa_obj('311', '12725', 'tcp', 'a.com')
    t_a312 = setup.create_tlsa_obj('312', '12725', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [],
            [t_a200, t_a201, t_a202, t_a210, t_a211, t_a212,
             t_a300, t_a301, t_a302, t_a310, t_a311, t_a312])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le


def test_config2():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.config2)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.ok

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))

    t_a1 = setup.create_tlsa_obj('201', '12725', 'tcp', 'a.com')
    t_a2 = setup.create_tlsa_obj('211', '12725', 'tcp', 'a.com')
    t_a3 = setup.create_tlsa_obj('301', '12725', 'tcp', 'a.com')
    t_a4 = setup.create_tlsa_obj('311', '12725', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [], [t_a1, t_a2, t_a3, t_a4])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le


def test_config6():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.config6)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.ok

    api0 = setup.create_api_binary_obj(str(s.bin / 'dns'))
    api1 = setup.create_api_binary_obj(['bin', '--flag1', 'input',
                                        "input with\t whitespace"])
    api2 = setup.create_api_c4_obj(zone='ZONE', email='me@domain.com',
                                   key='KEY')

    t_a1 = setup.create_tlsa_obj('201', '12725', 'tcp', 'a.com')
    t_a2 = setup.create_tlsa_obj('211', '12725', 'tcp', 'a.com')
    t_a3 = setup.create_tlsa_obj('301', '12725', 'tcp', 'a.com')
    t_a4 = setup.create_tlsa_obj('311', '12725', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api1, [], [t_a1, t_a2, t_a3, t_a4])

    t_b1 = setup.create_tlsa_obj('200', '1', 'sctp', 'b.com')
    t_b2 = setup.create_tlsa_obj('201', '1', 'sctp', 'W.com')
    t_b3 = setup.create_tlsa_obj('202', '1', 'tcp', 'X.com')
    t_b4 = setup.create_tlsa_obj('210', '1', 'sctp', 'Y.com')
    t_b5 = setup.create_tlsa_obj('211', '1', 'sctp', 'Z.com')
    t_b6 = setup.create_tlsa_obj('212', '1', 'sctp', 'A.com')
    t_b7 = setup.create_tlsa_obj('212', '1', 'udp', 'B.com')
    tb = setup.create_target_obj('b.com', api2, [],
                                 [t_b1, t_b2, t_b3, t_b4, t_b5, t_b6, t_b7])

    t_c1 = setup.create_tlsa_obj('200', '2', 'tcp', 'c.com')
    tc = setup.create_target_obj('c.com', api0, [], [t_c1])

    assert prog.target_list == [ta, tb, tc]
    assert prog.dane_directory == Path('/tmp/Q')
    assert prog.letsencrypt_directory == cwd / '../relative_path'



def test_fail_configX1():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX1)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    assert prog.target_list == []
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX2():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX2)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX3():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX3)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    assert prog.target_list == []
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX4():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX4)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    assert prog.target_list == []
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX5():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX5)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    assert prog.target_list == []
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX6():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX6)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX7():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX7)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX8():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX8)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX9():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX9)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX10():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX10)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX11():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX11)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX12():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX12)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX13():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX13)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX14():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX14)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX15():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX15)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX16():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX16)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', None, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX17():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX17)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX18():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX18)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', api, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == cwd / s.le


def test_fail_configX19():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX19)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    api = setup.create_api_binary_obj(str(s.bin / 'dns'))
    ta = setup.create_target_obj('a.com', api, [], [])

    assert prog.target_list == [ta]
    assert prog.dane_directory == Path('/tmp')
    assert prog.letsencrypt_directory == Path('/var/tmp')


def test_fail_configX20():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX20)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', None, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le


def test_fail_configX21():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX21)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', None, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le


def test_fail_configX22():
    s = setup.Init(keep=True)
    prog = setup.create_state_obj(s, config=s.configX22)
    cwd = Path.cwd()

    prog.log.init(prog.name, prog.version, prog.timenow)
    assert not prog.log.has_errors()

    retval = config.read(prog)
    assert retval == Prog.RetVal.config_failure

    t_a1 = setup.create_tlsa_obj('202', '1', 'tcp', 'a.com')
    ta = setup.create_target_obj('a.com', None, [], [t_a1])

    assert prog.target_list == [ta]
    assert prog.dane_directory == cwd / s.dane
    assert prog.letsencrypt_directory == cwd / s.le





import sys
from enum import Enum
import pathlib
import logging
import datetime
import fcntl

from alnitak import exceptions as Except
import alnitak


class State:
    """Internal program state.

    This class records data that is pertinent to the operation of the
    program. Virtuallly every function takes a State object as an
    argument since this is the central data store of things that function
    needs to operate properly.

    Class attributes are broadly classified into three groups. The first
    group is static data that controls how the program operates. The
    second group is data that controls how the program operates, but that
    can be changed with command-line flags. The last group is internal
    data that is set or read from config files or datafiles and needs to
    be shared between functions in order for them to operate.

    Attributes:
        name (str): name of the program.
        version (str): program version.
        copyright (str): copyright message.
        tlsa_parameters_regex (str): regex that specifies a valid TLSA
            parameter.
        tlsa_domain_regex (str): regex that specifies a valid TLSA domain
            name.
        tlsa_protocol_regex (str): regex that specifies a valid TLSA
            protocol.
        timenow (datetime.datetime): UTC time right now.
        ttl_min (int): the minimum allowed input to the '--ttl' flag.
        testing_mode (bool): normally 'False'. If set to 'True', then
            root-only processes are not run. This is just performing a
            chown on any files (e.g. the datafile).
        datafile (pathlib.Path): the datafile path.
        lockfile (pathlib.Path): the lock file path.
        can_lock (bool): whether the program should create a lock file.
        lock_fd (file object): the file object returned by 'open' when we
            open the lockfile.
        locked (bool): set to 'True' is the lock succeeded and the
            lockfile should be deleted at the end. If the program is run
            whilst another instance has acquired the lock, then this will
            still be 'False' and no deletion of the lockfile will take
            place: blind deletion will remove the lock when we don't want
            it to.

        config (pathlib.Path): the path of the config file.
        dane_directory (pathlib.Path): the path of the dane directory.
        letsencrypt_directory (pathlib.Path): the path of the Let's Encrypt
            directory (the directory the live and archive folders are in).
        letsencrypt_live_directory (pathlib.Path): the path of the live
            directory.
        ttl (int): the time-to-live value (in seconds). At least this
            number of seconds must pass since the publication of a new
            TLSA record before the old one is deleted.
        log (Log): an instance of the 'Log' class, which controls logging.
        recreate_dane (bool): set to 'True' if the '--reset' flag is
            given.

        args: the args given to argparse.
        target_list (list(Target)): list of targets in the config file.
        dane_domain_directories (dict(str: list(str))): for every folder
            in the live directory, set the key to the folder name (which
            will be a domain name). The value will be a list of symlinks
            in that folder (just the name of that symlink, not its path).
        renewed_domains list((str)): set to the value of the
            'RENEWED_DOMAINS' environment parameter, if set. Otherwise
            this will be set to an empty list.
        data (Data): the Data object that records the data lines read
            from (or need to be written to) a datafile.
    """
    def __init__(self, lock=True, testing=False):
        ## program constants
        self.name = "alnitak"
        self.version = alnitak.__version__
        self.copyright = "copyright (c) K. S. Kooner, 2019, MIT License"
        self.tlsa_parameters_regex = r"[23][01][012]"
        self.tlsa_domain_regex = r"((\w[a-zA-Z0-9-]*\w|\w+)\.)+\w+"
        self.tlsa_protocol_regex = r"\w+"
        self.timenow = datetime.datetime.utcnow()
        self.ttl_min = 0 # FIXME: leave me? Set to something like 60?
        self.testing_mode = testing
        self.datafile = ( pathlib.Path("/var")
                                    / self.name / str(self.name + ".data") )
        self.lockfile = pathlib.Path("/var/lock/{}.lock".format(self.name))
        self.can_lock = lock
        self.lock_fd = None
        self.locked = False

        ## program configuration data
        self.config = pathlib.Path("/etc/{}.conf".format(self.name))
        self.dane_directory = pathlib.Path("/etc/{}/dane".format(self.name))
        self.letsencrypt_directory = pathlib.Path("/etc/letsencrypt")
        self.letsencrypt_live_directory = self.letsencrypt_directory / "live"
        self.ttl = 86400
        self.log = Log(file="/var/log/{}.log".format(self.name), name=self.name)
        self.recreate_dane = False

        ## the following are data objects filled in during operation of the
        ## program
        self.args = None
        self.target_list = [ ]
        self.dane_domain_directories = { }
            # dictionary of keys that are dane domain subfolders, keyed
            # to a list of strings that are he symlinks in that subfolder.
            # For example:
            #    dane/
            #    |- x.com/
            #    |  |- link1@
            #    |  |- link2@
            #    |  |- link3@
            #    |  |- regfile1
            #    |  |- dir1/
            #    |
            #    |- y.com/
            #    |  |- link1@
            #    |  |- link2@
            #    |
            #    |- regfile1
            #
            # self.dane_domain_directories =
            #                   { 'x.com': [ 'link1', 'link2', 'link3' ],
            #                     'y.com': [ 'link1', 'link2' ] }
        self.renewed_domains = []
        self.data = Data()

    def lock(self):
        if not self.can_lock:
            return False
        try:
            self.lock_fd = open(str(self.lockfile), "w")
        except OSError as ex:
            raise Except.LockError(
                    "could not open lock file '{}': {}".format(
                                            ex.filename, ex.strerror.lower() ))
        try:
            fcntl.lockf(self.lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            return True
        self.locked = True
        return False

    def __del__(self):
        if self.can_lock and self.locked:
            try:
                self.lockfile.unlink()
            except OSError:
                # if file removal fails, we won't really care.
                pass

    def __str__(self):
        ret = "--- SYSTEM DATA ---------------\n"
        ret += "timenow: {0} {0:%s}\n".format(self.timenow)
        ret += "logfile: {}\n".format(self.log)
        ret += "ttl: {}\n".format(self.ttl)
        ret += "renewed_domains: {}\n".format(self.renewed_domains)
        ret += "datafile: {}\n".format(self.datafile)
        ret += "dane: {}\n".format(self.dane_directory)
        ret += "letsencrypt: {}\n".format(self.letsencrypt_directory)
        ret += "live: {}\n".format(self.letsencrypt_live_directory)
        ret += "dane_domain_directories: {}".format(
                                                self.dane_domain_directories)
        for t in self.target_list:
            ret += "\n------------------\n{}".format(t)
        ret += "\n------------------\n"

        ret += "\n------------------\ndatalines:"
        #for l in self.datafile_lines:
        #    ret += "\n    =============\n{}".format(l)

        ret += "{}\n".format(self.data)

        return ret

    def init_logging(self, args):
        self.args = args
        self.log.init(self.name, self.version, self.timenow)

    def make_absolute(self, path):
        p = pathlib.Path(path)
        if p.is_absolute():
            return p
        return pathlib.Path.cwd() / p

    def set_letsencrypt_directory(self, path):
        self.letsencrypt_directory = self.make_absolute(path)
        self.letsencrypt_live_directory = self.letsencrypt_directory / "live"

    def set_dane_directory(self, path):
        self.dane_directory = self.make_absolute(path)

    def set_ttl(self, ttl):
        self.ttl = ttl

    def set_config_file(self, path):
        self.config = self.make_absolute(path)


class LogType(Enum):
    """Where to log to, if at all."""
    logfile = 0
    stdout = 1
    no = 3

class LogLevel(Enum):
    """Logging level."""
    nolog = 0
    normal = 1
    verbose = 2
    full = 3

class Log():
    """Class to control logging of information.

    Logging schenarios:
    1.  flags: <NONE>        info   -> logfile
                             errors -> logfile, stderr
    2.  flags: -l-           info   -> stdout
                             errors -> stderr
    3.  flags: -lno          info   ->X
                             errors -> stderr
    4.  flags: -q            info   -> logfile
                             errors -> logfile
    5.  flags: -l- -q        info   ->X
    6.  flags: -lno -q       errors ->X

    7.  flags: -Lno          info   ->X
                             errors -> logfile, stderr
    8.  flags: -l- -Lno      info   ->X
                             errors -> stderr
    9.  flags: -lno -Lno     info   ->X
                             errors -> stderr
    10. flags: -q -Lno       info   ->X
                             errors -> logfile
    11. flags: -l- -q -Lno   info   ->X
    12. flags: -lno -q -Lno  errors ->X

    Attributes:
        type (LogType): whether we are logging to a file, stdout or not
            logging at all.
        file (pathlib.Path): the logfile.
        level (LogLevel): the level of info to log.
        quiet (bool): whether the '--quiet' flag was given or not. If set
            to 'True', then absolutely no messages are printed, not even
            error messages.
        log_info: the Logger for info messages.
        log_err: the Logger for error messages.
        logfile_checked (bool): when we write to a logfile, we need to do
            some checks on the file. If set to 'True', we don't bother
            running the checks again. We also use this bool to tell if we
            are writing to a logfile since it is only ever 'True' if
            writing to a logfile has been requested (and hence, a check
            was requested in the first place).
        error_handler_available (bool): set to 'True' if there is an
            available output for error messages. This will only be set if
            both: the logging flags to the program made available an
            error stream AND at least one stream was initiaized without
            errors. Note: a value of 'True' does NOT mean all the error
            streams were initialized successfully: only that at least one
            was.
        error_msg (list(str)): error messages for output streams that
            failed initialization. The reason why these are stored and not
            just printed to the screen is precisely because some streams
            failed to initialize: where we need to print these error
            messages (if we can do so at all) depends on what streams are
            actually availabale. The streams that have failed are marked
            in the following fail_X attributes...
        fail_error_logfile (bool): set to 'True' if the error stream to
            the logfile was not initialized without errors.
        fail_error_stderr (bool): set to 'True' if the error stream to
            stderr was not initialized without errors.
        fail_info_logfile (bool): set to 'True' if the info stream to
            the logfile was not initialized without errors.
        fail_info_stdout (bool): set to 'True' if the info stream to
            stdout was not initialized without errors.
        fail_output (bool): set to 'True' if printing to a stream (info
            or error) encountered an error AFTER it was successfully
            initialized. When this occurs, we simply print an error to
            the error stream, if one is available (c.f. the
            'error_handler_available' attribute).
    """
    def __init__(self, name, file):
        self.type = LogType.logfile
        self.file = pathlib.Path(file)
        self.level = LogLevel.normal
        self.quiet = False

        self.log_info = logging.getLogger("{}:out".format(name))
        self.log_err = logging.getLogger("{}:err".format(name))
        self.log_info.setLevel(logging.INFO)
        self.log_err.setLevel(logging.INFO)
        self.logfile_checked = False

        self.error_handler_available = False
        self.error_msg = []
        self.fail_error_logfile = False
        self.fail_error_stderr = False
        self.fail_info_logfile = False
        self.fail_info_stdout = False
        self.fail_output = False

    def __str__(self):
        if self.type == LogType.logfile:
            return "[{}] [q:{}] {}".format(self.level, self.quiet, self.file)
        elif self.ttpe == LogType.stdout:
            return "[{}] [q:{}] {}".format(self.level, self.quiet, self.type)
        else:
            return "{}".format(self.type)

    def set_file(self, file):
        self.type = LogType.logfile
        self.file = pathlib.Path(file)

    def set_stdout(self):
        self.type = LogType.stdout
        self.file = sys.stdout

    def set_nolog(self):
        self.type = LogType.no
        self.file = None

    def set_level(self, level):
        self.level = level

    def set_quiet(self):
        self.quiet = True

    def has_errors(self):
        return (self.fail_output or self.fail_error_logfile
                or self.fail_error_stderr or self.fail_info_logfile
                or self.fail_info_stdout)

    def init(self, progname, progversion, timenow):
        """Initialize the logging handlers."""

        # set the error handler:
        if self.quiet:
            if self.type == LogType.logfile:
                self.error_handler_file(progname)
        else:
            self.error_handler_stderr()
            if self.type == LogType.logfile:
                self.error_handler_file(progname)

        # set the info handler:
        if self.quiet:
            if self.type == LogType.logfile:
                self.info_handler_file(progname)
        else:
            if self.type == LogType.logfile:
                self.info_handler_file(progname)
            elif self.type == LogType.stdout:
                self.info_handler_stdout()

        # now, if any of the handlers above have failed, then we need to
        # try to print an error message.
        if (not self.fail_info_logfile and not self.fail_info_stdout
            and self.logfile_checked):
                self.write_header(progname, progversion, timenow)

        if (self.fail_error_logfile or self.fail_error_stderr
                or self.fail_info_logfile or self.fail_info_stdout):
            if self.error_handler_available:
                self.error(self.error_msg)
                self.error("logging may be incomplete")


    def write_header(self, progname, progversion, timenow):
        # create header section for the logfile
        arg_str = sys.argv[0]
        for a in sys.argv[1:]:
            if ' ' in a or '\t' in a or '\n' in a:
                arg_str += " '{}'".format(a)
            else:
                arg_str += " {}".format(a)
        self.info1("---------------------------------------------\n{0} {1}\n{2} ({2:%s})\nprogram run as: {3}\n---------------------------------------------".format(progname, progversion, timenow, arg_str))


    def info_handler_file(self, progname):
        if self.check_logfile(progname):
            self.fail_info_logfile = True
            return True
        try:
            handler = logging.FileHandler(str(self.file))
            handler.setFormatter( logging.Formatter('%(message)s') )
            self.log_info.addHandler(handler)
        except OSError as ex:
            self.error_msg += [
                    "creating handler to logfile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()) ]
            self.fail_info_logfile = True
            return True
        return False


    def info_handler_stdout(self):
        try:
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter( logging.Formatter('%(message)s') )
            self.log_info.addHandler(handler)
        except OSError as ex:
            self.error_msg += [
                    "creating handler to stdout failed: {}".format(
                                                        ex.strerror.lower()) ]
            self.fail_info_stdout = True
            return True
        return False


    def error_handler_file(self, progname):
        if self.check_logfile(progname):
            self.fail_error_logfile = True
            return True
        try:
            handler = logging.FileHandler(str(self.file))
            handler.setFormatter( logging.Formatter('%(message)s') )
            self.log_err.addHandler(handler)
        except OSError as ex:
            self.error_msg += [
                    "creating handler to logfile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()) ]
            self.fail_error_logfile = True
            return True

        self.error_handler_available = True
        return False


    def error_handler_stderr(self):
        try:
            handler = logging.StreamHandler(sys.stderr)
            handler.setFormatter( logging.Formatter('%(message)s') )
            self.log_err.addHandler(handler)
        except OSError as ex:
            self.error_msg += [
                    "creating handler to stderr failed: {}".format(
                                                        ex.strerror.lower()) ]
            self.fail_error_stderr = True
            return True

        self.error_handler_available = True
        return False


    def check_logfile(self, progname):
        if self.logfile_checked:
            return False
        self.logfile_checked = True

        if self.file.is_dir():
            self.file = self.file / "{}.log".format(progname)

        try:
            self.file = self.file.resolve()
        except FileNotFoundError as ex:
            pass
            # we'll catch the parent directory not existing next. We don't
            # want to exit here if the log file itself doesn't exist.
        except RuntimeError as ex:
            self.error_msg += [
                    "log file '{}': file could not be resolved".format(
                                                                ex.filename) ]
            return True

        if not self.file.parent.exists():
            self.error_msg += [ "logging: directory '{}' not found".format(
                                                            self.file.parent) ]
            return True

        return False


    def printmsg(self, msg, level):
        try:
            if self.quiet and self.type == LogType.logfile:
                if level.value <= self.level.value:
                    self.log_info.info(msg)
            elif self.type != LogType.no:
                if level.value <= self.level.value:
                    self.log_info.info(msg)
        except OSError as ex:
            self.fail_output = True
            if self.error_handler_available:
                self.error("writing log info failed: {}".format(
                                                    ex.strerror.lower()) )

    def info1(self, msg):
        self.printmsg(msg, LogLevel.normal)

    def info2(self, msg):
        self.printmsg(msg, LogLevel.verbose)

    def info3(self, msg):
        self.printmsg(msg, LogLevel.full)

    def error(self, msg):
        if isinstance(msg, str):
            lines = msg.splitlines()
        else:
            lines = msg
        for l in lines:
            try:
                if self.quiet:
                    if self.type == LogType.logfile:
                        self.log_err.info("error: {}".format(l))
                else:
                    self.log_err.info("error: {}".format(l))
            except OSError:
                self.fail_output = True


    def warning(self, msg):
        if isinstance(msg, str):
            lines = msg.splitlines()
        else:
            lines = msg
        for l in lines:
            try:
                if self.quiet:
                    if self.type == LogType.logfile:
                        self.log_err.info("warning: {}".format(l))
                else:
                    self.log_err.info("warning: {}".format(l))
            except OSError:
                self.fail_output = True




class RetVal(Enum):
    """Exit code values."""
    ok = 0
    exit_ok = 256
    exit_failure = 1
    continue_failure = 257
    config_failure = 3



class Api:
    """API scheme base class.

    Attributes:
        type (ApiType): the specific API scheme.
    """

    def __init__(self, type):
        self.type = type

    def __eq__(self, a):
        return self.type == a.type

class ApiCloudflare4(Api):
    """The Cloudflare4 API scheme.

    Attributes:
        zone (str): the Cloudflare zone.
        email (str): the email of the user to login as.
        key (str): the key of the user to login with.
    """

    def __init__(self):
        super().__init__(ApiType.cloudflare4)
        self.zone = None
        self.email = None
        self.key = None

    def __str__(self):
        return "    - {}\n       zone: ...({})\n       email: ...({})\n       key: ...({})".format(self.type, len(self.zone), len(self.email), len(self.key))

    def __eq__(self, a):
        return (self.type == a.type and self.zone == a.zone
                and self.email == a.email and self.key == a.key)

class ApiBinary(Api):
    """The 'binary' API scheme

    Attributes:
        command (list(str)): the command to run (and any flags/inputs).
        uid (int): the UID to run the command as. If set to 'None' run
            as the same user as the calling user (usually 'root').
        gid (int): NOT USED. Envisioned as the GID to run the process
            under. This is instead set from the passwd info of the UID.
    """

    def __init__(self, command, uid=None, gid=None):
        super().__init__(ApiType.binary)
        self.command = command
        self.uid = uid
        self.gid = gid # NOTE: not used.

    def __str__(self):
        return "    - {}\n       command: {} [uid: {}]".format(self.type, self.command, self.uid, self.gid)

    def rstr(self):
        return "[{}]".format("] [".join(self.command)) + " (uid:{})".format(self.uid, self.gid)

    def __eq__(self, a):
        return (self.type == a.type and self.command == a.command
                and self.uid == a.uid and self.gid == a.gid)


class ApiType(Enum):
    """The API scheme."""
    cloudflare4 = 0
    binary = 1



class Tlsa:
    """Class recording the data of a TLSA record.

    Attributes:
        usage (str): either '2' or '3'.
        selector (str): either '0' or '1'.
        matching (str): either '0', '1' or '2'.
        port (str): port number of the TLSA record.
        protocol (str): protocol of the TLSA record.
        domain (str): domain of the TLSA record.
        publish (bool): whether to publish the record (and create a
            posthook line) or not. Normally this is always 'True'. This
            is only ever 'False' when we have re-renewed a certificate
            whose hash to publish has not changed. In this case we do not
            need to publish a TLSA and add a posthook line since this has
            all already been done.
    """

    def __init__(self, param, port, protocol, domain):
        self.usage = param[0]
        self.selector = param[1]
        self.matching = param[2]
        self.port = port
        self.protocol = protocol
        self.domain = domain
        self.publish = True

    def __eq__(self, t):
        return ( self.usage == t.usage and self.selector == t.selector
                    and self.matching == t.matching and self.port == t.port
                    and self.protocol == t.protocol
                    and self.domain == t.domain and self.publish == t.publish )

    def __str__(self):
        return "    - tlsa: {}{}{} {} {} {} [state:{}]".format(
                    self.usage, self.selector, self.matching, self.port,
                    self.protocol, self.domain, self.publish)

    def pstr(self):
        return "{}{}{} {} {} {}".format(
                    self.usage, self.selector, self.matching, self.port,
                    self.protocol, self.domain)

    def publish_off(self):
        self.publish = False


class Cert:
    """A set of corresponding live, archive and dane certificates.

    Attributes:
        dane (pathlib.Path): the absolute path of the dane certificate.
        live (pathlib.Path): the absolute path of the live certificate.
        archive (pathlib.Path): the absolute path of the archive certificate.
    """

    def __init__(self, dane, live, archive):
        self.dane = pathlib.Path(dane)
        self.live = pathlib.Path(live)
        self.archive = pathlib.Path(archive)

    def __eq__(self, c):
        return ( self.dane == c.dane and self.live == c.live
                    and self.archive == c.archive )

    def __str__(self):
        return "    - dane: {}\n      . live: {}\n      . archive: {}".format(
                                            self.dane, self.live, self.archive)

    def pstr(self):
        return "  + dane: {}\n  + live: {}\n  + archive: {}".format(
                                            self.dane, self.live, self.archive)


class Target:
    """A config file target.

    Attributes:
        domain (str): this is the section of the config file target. Its
            value should be the name of a folder in the archive directory.
        certs (list(Cert)): a list of sets of dane, live, archive certs.
        tlsa (list(Tlsa)): a list of Tlsa records.
        api (Api): A derived class of Api what stores the API scheme of
            the target.
    """

    def __init__(self, domain):
        self.domain = domain        # This is the subfolder 
        self.certs = []             # [ Cert()... ]
        self.tlsa = []              # [ Tlsa()... ]
        self.api = None             # Api()

    def __str__(self):
        #ret = "{}\n".format(self.domain)
        #for c in self.certs:
        #    ret += "{}\n    ~~~~~~~~~~~\n".format(c)
        #for t in self.tlsa:
        #    ret += "  tlsa: {}\n".format(t)
        #ret += "  api: {}".format(self.api)
        #return ret
        ret = "  + {}\n".format(self.domain)
        for c in self.certs:
            ret += "{}\n".format(c)
        for t in self.tlsa:
            ret += "{}\n".format(t)
        ret += "{}".format(self.api)
        return ret

    def __eq__(self, t):
        return (self.domain == t.domain and self.certs == t.certs
                and self.tlsa == t.tlsa and self.api == t.api)

    def matches_domain(self, str):
        if str == self.domain:
            return True
        return False

    def add_tlsa(self, t):
        self.tlsa += [ t ]

    def set_api(self, type, data):
        self.api = Api(type, data)

    def add_cert(self, dane, live, archive):
        self.certs += [ Cert(dane, live, archive) ]


class ConfigState:
    """Class to record syntax errors in the config file.

    Attributes:
        linepos (int): the line the error is on.
        errors (int): the number of errors encountered.
    """

    def __init__(self):
        self.linepos = None
        self.errors = 0

    def line(self, linepos):
        self.linepos = linepos

    def add_error(self, prog, msg):
        self.errors += 1
        if self.linepos:
            prog.log.error("config file: line {}: {}".format(self.linepos, msg))
        else:
            prog.log.error("config file: {}".format(msg))



class DataLine:
    """Base class for datalines.

    Attributes:
        type (DataLineType): the type of the line.
        domain (str): the domain of the line (e.g. 'example.com').
        lineno (int): the line number in the datafile the line was on.
            Any new lines to be written are given a line number of zero.
        state (DataLineState): whether to write the line to a new datafile
            or not.
    """

    def __init__(self, type, domain, lineno):
        self.type = type
        self.domain = domain
        self.lineno = lineno
        self.state = DataLineState.write

    def write_state_off(self):
        self.state = DataLineState.skip

class DataPre(DataLine):
    """Class recording the data in a datafile posthook line.

    Attributes:
        cert (Cert): records the dane, live and archive certificates.
        pending (str): Either '0' when written before any posthook
            operation has been made (i.e., no posthook lines also
            present), or else '1' when posthook lines also present.
    """

    def __init__(self, domain, lineno, dane, live, archive, pending):
        super().__init__(DataLineType.pre, domain, lineno)
        self.cert = Cert(dane, live, archive)
        self.pending = pending

    def __eq__(self, l):
        return ( self.type == l.type and self.domain == l.domain
                    and self.cert == l.cert and self.pending == l.pending
                    and self.state == l.state )

    def __str__(self):
        return "  + type: {}\n  + domain: {}\n  + line: {}\n{}\n  + pending: {}\n  + state: {}".format(self.type, self.domain, self.lineno, self.cert.pstr(), self.pending, self.state)

    def is_strict_eq(self, l):
        return (self.type == l.type and self.domain == l.domain
                and self.lineno == l.lineno and self.state == l.state
                and self.cert == l.cert and self.pending == l.pending)

    def pending_on(self):
        self.pending = '1'

    def pending_off(self):
        self.pending = '0'

class DataPost(DataLine):
    """Class recording the data in a datafile posthook line.

    Attributes:
        tlsa (Tlsa): TLSA record, constructed from the data line.
        pending (str): Either '0' if the TLSA record above was published,
            or else '1' if publication failed and still needs to be done.
        time (str): seconds in unix time. This is the time when the
            TLSA record was published, or else when the first attempt to
            do so was made.
        hash (str): the 'certificate data' of the record to publish or
            was published. This will be the new live certificate after
            renewal.
        mark_delete (bool): if the record above should be deleted.
            Deletion should be done after any records are published, so
            we need to mark a record for deletion before we actually do
            it.
    """

    def __init__(self, domain, lineno, tlsa, pending, time, hash):
        super().__init__(DataLineType.post, domain, lineno)
        self.tlsa = tlsa
        self.pending = pending
        self.time = time
        self.hash = hash
        self.mark_delete = False

    def __eq__(self, l):
        return ( self.type == l.type and self.domain == l.domain
                    and self.tlsa == l.tlsa and self.pending == l.pending
                    and self.hash == l.hash and self.state == l.state )

    def __str__(self):
        return "  + type: {}\n  + domain: {}\n  + line: {}\n  + tlsa: {}\n  + pending: {}\n  + time: {}\n  + hash: {}\n  + state: {}".format(self.type, self.domain, self.lineno, self.tlsa.pstr(), self.pending, self.time, self.hash, self.state)

    def is_strict_eq(self, l):
        return (self.type == l.type and self.domain == l.domain
                and self.lineno == l.lineno and self.state == l.state
                and self.tlsa == l.tlsa and self.pending == l.pending
                and self.hash == l.hash and sel.make_absolute == l.mark_delete)

    def pending_off(self):
        self.pending = '0'

    def change_time(self, time):
        self.time = time

    def mark_for_deletion(self):
        self.mark_delete = True

class DataDelete(DataLine):
    """Class recording the data in a datafile delete line.

    Attributes:
        tlsa (Tlsa): TLSA record, constructed from the data line.
        count (str): number of attempt to delete the TLSA record above.
            Starts from '1' since the failed delete that caused the delete
            line to be made counts as the first attempt. Incremented for
            every failed delete or else the line is deleted.
        time (str): seconds in unix time.
        hash (str): the 'certificate data' of the TLSA record to delete.
    """

    def __init__(self, domain, lineno, tlsa, count, time, hash):
        super().__init__(DataLineType.delete, domain, lineno)
        self.tlsa = tlsa
        self.count = count
        self.time = time
        self.hash = hash

    def __eq__(self, l):
        return ( self.type == l.type and self.domain == l.domain
                    and self.tlsa == l.tlsa and self.count == l.count
                    and self.hash == l.hash and self.state == l.state )

    def __str__(self):
        return "  + type: {}\n  + domain: {}\n  + line: {}\n  + tlsa: {}\n  + count: {}\n  + time: {}\n  + hash: {}\n  + state: {}".format(self.type, self.domain, self.lineno, self.tlsa.pstr(), self.count, self.time, self.hash, self.state)

    def is_strict_eq(self, l):
        return (self.type == l.type and self.domain == l.domain
                and self.lineno == l.lineno and self.state == l.state
                and self.tlsa == l.tlsa and self.count == l.count
                and self.hash == l.hash)

    def increment_count(self):
        c = int(self.count)
        c += 1
        self.count = str(c)

class DataLineType(Enum):
    """Type of a datafile 'data line'."""
    pre = 0
    post = 1
    delete = 2

class DataLineState(Enum):
    """Dataline state is whether the line should be written or not."""
    write = 0
    skip = 1

class DataGroup:
    """A group of datalines is a set of lines that have a common domain.

    Mathematicians: yes, set not group...

    Attributes:
        domain (str): the common domain of all the data lines.
        target (Target): the target (from the config file) that applies to
            the domain above.
        pre: (list(DataPre)): list of prehook lines.
        post: (list(DataPost)): list of posthook lines.
        special: (list(DataDelete)): list of delete lines.
    """

    def __init__(self, prog, line):
        self.domain = line.domain

        for t in prog.target_list:
            if line.domain == t.domain:
                self.target = t
                break
        else:
            self.target = None
            prog.log.warning(
                    "line {}: domain '{}' not found in config file".format(
                                                    line.lineno, line.domain))
        self.pre = []
        self.post = []
        self.special = []
        if line.type == DataLineType.pre:
            self.add_pre(line)
        elif line.type == DataLineType.post:
            self.add_post(line)
        else:
            self.add_special(line)

    def __str__(self):
        ret = " ++ group '{}'\n{}".format(self.domain, self.target)
        for l in self.post:
            ret += "\n    --- line ---------\n{}".format(l)
        for l in self.pre:
            ret += "\n    --- line ---------\n{}".format(l)
        for l in self.special:
            ret += "\n    --- line ---------\n{}".format(l)
        return ret

    def add_line(self, line):
        if line.type == DataLineType.pre:
            self.add_pre(line)
        elif line.type == DataLineType.post:
            self.add_post(line)
        else:
            self.add_special(line)

    def add_pre(self, line):
        if line in self.pre:
            return
        self.pre += [ line ]

    def add_post(self, line):
        if line in self.post:
            return
        self.post += [ line ]

    def add_special(self, line):
        if line in self.special:
            return
        self.special += [ line ]


class Data:
    """Class to record all the data in the datafile.

    Attributes:
        groups (list(DataGroup)): Lines in the datafile are grouped by
            their domain.
    """

    def __init__(self):
        self.groups = []

    def add_line(self, prog, line):
        """Line is added to either an existing group, or a new group."""
        for d in self.groups:
            if d.domain == line.domain:
                d.add_line(line)
                break
        else:
            self.groups += [ DataGroup(prog, line) ]

    def __str__(self):
        ret = " ++ printing data groups:"
        for g in self.groups:
            ret += "\n{}".format(g)
        return ret






import re
from codecs import encode
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization

from alnitak import exceptions as Except

# Note: python 3.5+ can use X.hex() instead of encode(X,'hex').decode('ascii').
# If going to change that, then remove the import above.

def get_live(tlsa, pre):
    """Return a live certificate name from one of the prehook lines.

    Return a name (pathlib.Path) of a live certificate from one of the
    prehook lines, the appropriate certificate depending on the DANE
    record requested.

    Args:
        tlsa (Tlsa): not changed.
        pre (DataPre): not changed.

    Returns:
        pathlib.Path: taken from 'pre'. If no certificate is found an
            exception is raised.

    Raises:
        InternalError: there ought to be a prehook line for every pem cert
            in the domain folder, so there ought to be a certificate for
            whatever DANE record is requested.
    """
    if tlsa.usage == '2':
        name_list = [ "chain.pem", "fullchain.pem" ]
    else:
        name_list = [ "cert.pem", "fullchain.pem" ]

    for cert in [ l.cert.live for l in pre ]:
        if cert.name in name_list:
            return cert

    raise Except.InternalError("suitable cert could not be found")


def get_archive(tlsa, pre):
    """Return an archive certificate name from one of the prehook lines.

    Return a name (pathlib.Path) of an archive certificate from one of
    the prehook lines, the appropriate certificate depending on the DANE
    record requested.

    Args:
        tlsa (Tlsa): not changed.
        pre (DataPre): not changed.

    Returns:
        pathlib.Path: taken from 'pre'. If no certificate is found an
            exception is raised.

    Raises:
        InternalError: there ought to be a prehook line for every pem cert
            in the domain folder, so there ought to be a certificate for
            whatever DANE record is requested.
    """
    if tlsa.usage == '2':
        name_regex = r"(full)?chain[0-9]+\.pem"
    else:
        name_regex = r"(cert|fullchain)[0-9]+\.pem"

    for l in pre:
        if re.match(name_regex, l.cert.archive.name):
            return l.cert.archive

    raise Except.InternalError("suitable cert could not be found")

def read_cert(cert, tlsa):
    """Return the PEM-encoded content of a certificate file.

    Open a certificate file and return a str that is the PEM-encoded
    public key info in that file, in the case of a fullchain file, only
    the necessary key is returned, depending on the DANe record requested.

    Args:
        cert (pathlib.Path): PEM-encoded file to read.

    Returns:
        str: PEM-encoded content of 'cert' (with BEGIN and END headers and
            footers). If 'cert' is a fullchain file, the relevent
            certificate PEM is extracted.

    Raises:
        DNSProcessingError: if 'cert' could not be opened or no appropriate
            PEM section in a fullchain file could be found. Also returned
            if there is no content of the 'cert' file.
    """
    try:
        with open(str(cert), "r") as file:
            cert_data = file.read()
    except FileNotFoundError as ex:
        raise Except.DNSProcessingError(
                "creating hash: certificate file '{}' not found".format(
                                                                ex.filename))
    except OSError as ex:
        raise Except.DNSProcessingError(
                "creating hash: certificate file '{}': {}".format(
                                        ex.filename, ex.strerror.lower()))

    if re.match(r"fullchain[0-9]*\.pem$", cert.name):
        pems = []
        temp = ""
        for l in cert_data.splitlines():
            if l == "-----BEGIN CERTIFICATE-----":
                if temp:
                    pems += [ temp ]
                temp = l
            else:
                temp += "\n{}".format(l)
        if temp:
            pems += [ temp ]

        if tlsa.usage == '2':
            if len(pems) < 2:
                raise Except.DNSProcessingError("creating hash: '{}' file: no intermediate certificate found".format(cert))
            cert_data = pems[1]
        else:
            if len(pems) < 1:
                raise Except.DNSProcessingError(
                    "creating hash: '{}' file: no certificate found".format(
                                                                        cert))
            cert_data = pems[0]

    if len(cert_data) == 0:
        raise Except.DNSProcessingError(
                "creating hash: '{}' file: no data found".format(cert))

    return cert_data

def get_hash(selector, matching, data):
    """Create a DANE record 'certificate data' (hash) string.

    Given a certificate PEM-encoded public key, generate 'certificate data'
    (which we otherwise call a 'hash') for the DANE record.

    Args:
        selector (str): TLSA selector field (0|1).
        matching (str): TLSA matching-type field (0|1|2).
        data (str): PEM-encoded public key data.

    Returns:
        str: the 'certificate data' (hash).

    Raises:
        InternalError: if generating the 'certificate data' fails for any
            reason.
    """
    cert = x509.load_pem_x509_certificate(
                                    bytes(data, 'utf-8'), default_backend())

    try:
        if selector == '0':
            # use the full certificate
            if matching == '0':
                return encode(cert.public_bytes(
                                    encoding=serialization.Encoding.DER),
                                    'hex').decode('ascii')
            elif matching == '1':
                return encode(cert.fingerprint(hashes.SHA256()),
                              'hex').decode('ascii')
            else:
                return encode(cert.fingerprint(hashes.SHA512()),
                              'hex').decode('ascii')
        else:
            # use the public key
            der = cert.public_key().public_bytes(
                        encoding=serialization.Encoding.DER,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo)
            if matching == '0':
                return encode(der, 'hex').decode('ascii')
            elif matching == '1':
                digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
                digest.update(der)
                return encode(digest.finalize(), 'hex').decode('ascii')
            else:
                digest = hashes.Hash(hashes.SHA512(), backend=default_backend())
                digest.update(der)
                return encode(digest.finalize(), 'hex').decode('ascii')
    except cryptography.exceptions.UnsupportedAlgorithm:
        raise Except.InternalError("unsupported hash algorithm")
    except TypeError:
        raise Except.InternalError("certificate data not a byte string")







class ConfigError(RuntimeError):
    """Configuration file syntax errors."""
    def __init__(self, message=None):
        self.message = message

class LockError(RuntimeError):
    """Lock file errors."""
    def __init__(self, message=None):
        self.message = message

class InternalError(RuntimeError):
    """Internal errors."""
    def __init__(self, message=None):
        self.message = message


class DNSExcept(RuntimeError):
    """Base class for exceptions raised in processing data."""
    def __init__(self, message=None):
        self.message = message

class DNSSkip(DNSExcept):
    """Exceptions raised in processing data that are not errors."""
    pass

class DNSError(DNSExcept):
    """Exceptions raised by errors in processing data."""
    pass


class DNSNotLive(DNSSkip):
    """Raised when DNS records are not yet up (not yet 'live')."""
    pass

class DNSSkipProcessing(DNSSkip):
    """Raised when further processing should be skipped."""
    pass


class DNSProcessingError(DNSError):
    """Raised when processing has encountered an error."""
    pass

class DNSNoReturnError(DNSError):
    """Error that should not cause the program to exit non-zero."""
    pass


class PrivError(RuntimeError):
    """Exception for errors raised when dropping privileges."""
    def __init__(self, message=None):
        self.message = message




import sys
import os
import re
import argparse
import pathlib
import logging

from alnitak import prog as Prog
from alnitak import exceptions as Except
from alnitak.api import cloudflare4
from alnitak.api import binary
from alnitak import certop
from alnitak import config
from alnitak import datafile



def init_dane_directory(prog):
    """Create the dane directory and dane domain subdirectories.

    This function will check to see if the dane directory is 'sane' and
    will create the necessary folders and symlinks if they are missing.
    It will not try to fix mistakes unless 'prog.recreate_dane' is set
    to 'True': it will exit if errors are encountered.
    Note that even when 'prog.recreate_dane' is set to 'True', we will
    not remove files or folders that are no longer used (e.g. if a
    previous config file defined the domain 'x.com' and this program
    created a directory 'x.com' in the dane directory previously, if the
    config file now replaces 'x.com' with 'y.com', we will certainly
    create a new directory called 'y.com', but we will NOT remove 'x.com').
    If a completely clean directory is required, the user should manually
    remove the dane directory and then call this program to create a new
    one from scratch.

    Args:
        prog (State): program internal state.

    Returns:
        RetVal: returns 'RetVal.exit_failure' if any errors are
            encountered, and 'RetVal.ok' otherwise.
    """
    prog.log.info2("+++ initializing 'dane' direcory: '{}'".format(
                                                        prog.dane_directory))

    # create dane directory if it doesn't exist. We'll restrict the
    # permissions of this file to 700 like the letsencrypt directories do.
    try:
        prog.dane_directory.mkdir(mode=0o700, parents=True)
    except FileExistsError as ex:
        # Note: this catch can be removed for python 3.5+ since mkdir()
        # accepts the 'exist_ok' parameter.
        if prog.dane_directory.is_dir():
            # if directory exists, that is fine
            pass
        else:
            prog.log.error(
                "creating dane directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
            return Prog.RetVal.exit_failure
    except OSError as ex:
        prog.log.error(
            "creating dane directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure


    # if the dane directory alrwady exists, then its permissions might not
    # be correct. If told to recreate the directory, fix its permissions now:
    if prog.recreate_dane:
        prog.log.info2(
                " ++ checking/fixing mode of dane direcory: should be '0700'")
        try:
            prog.dane_directory.chmod(0o700)
        except OSError as ex:
            prog.log.error("changing permissions of dane directory '{}' failed: {}".format(ex.filename, ex.strerror.lower()))
            return Prog.RetVal.exit_failure

        prog.log.info2(
            " ++ checking/fixing owner of dane direcory: should be 'root:root'")
        try:
            # when running tests (unless run as root) we won't have
            # permissions to do this, so we have to skip this when testing
            if not prog.testing_mode:
                os.chown(str(prog.dane_directory), 0, 0)
        except OSError as ex:
            prog.log.error(
                    "changing owner of dane directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
            return Prog.RetVal.exit_failure


    # get a list of directories in the letsencrypt live directory (named after
    # domains)
    prog.log.info3("  + domain directories in the live direcory '{}':".format(
                                            prog.letsencrypt_live_directory))
    try:
        live_domains = [ pathlib.Path(prog.letsencrypt_live_directory / f)
                            for f in os.listdir(
                                        str(prog.letsencrypt_live_directory)) ]
    except OSError as ex:
        prog.log.error(
                "letsencrypt live directory '{}': {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure
    for d in live_domains:
        prog.log.info3("    - {}".format(d))


    retval = Prog.RetVal.ok

    # for each domain folder found in the list 'live_domains' populated above,
    # we need to do 2 things:
    #   1) create an identically-named folder in the dane directory
    #   2) populate that directory with symlinks to the _symlinks_ in the
    #      live directory
    for d in live_domains:

        dane_d = pathlib.Path(prog.dane_directory / d.name)

        prog.log.info2(" ++ checking dane domain directory '{}'".format(dane_d))

        # implement 1): create dane/$(basename d)
        try:
            dane_d.mkdir()
        except FileExistsError as ex:
            # Note: this catch can be removed for python 3.5+ since mkdir()
            # accepts the 'exist_ok' parameter.
            if prog.dane_directory.is_dir():
                # if directory exists, that is fine
                pass
            else:
                prog.log.error(
                    "creating dane domain directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
                retval = Prog.RetVal.exit_failure
                continue
        except OSError as ex:
            prog.log.error(
                    "creating dane domain directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
            retval = Prog.RetVal.exit_failure
            continue

        # implement 2): first, we get a list of symlinks in the live domain
        # directory
        prog.log.info3(
            "  + creating symlinks to live domain symlinks...".format(dane_d))
        try:
            link_list = [ f for f in os.listdir(str(d))
                                        if pathlib.Path(d / f).is_symlink() ]
        except OSError as ex:
            prog.log.error(
                    "letsencrypt (live) domain directory '{}': {}".format(
                                            ex.filename, ex.strerror.lower()))
            retval = Prog.RetVal.exit_failure
            continue


        # let's add the domain directory and the symlinks inside it to
        # prog.dane_domain_directories
        prog.dane_domain_directories[d.name] = link_list


        # implement 2): ...then, we create symlinks in the dane domain
        # directory, pointing to the live domain symlinks
        for l in link_list:
            prog.log.info3("    - {}".format(l))

            dane_l = pathlib.Path(dane_d / l)

            # if recreating the dane directory, remove the file even if it is
            # OK and we'll just create a new symlink.
            try:
                if prog.recreate_dane:
                    dane_l.unlink()
            except FileNotFoundError:
                # file doesn't exist is fine
                pass
            except OSError as ex:
                prog.log.error("removing symlink '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
                retval = Prog.RetVal.exit_failure
                continue


            try:
                dane_l.symlink_to( relative_to(dane_l, d / l) )
            except FileExistsError as ex:
                # when recreating the directory, the file _should not_
                # exist...
                if prog.recreate_dane:
                    prog.log.error(
                            "recreating symlink '{}': file exists".format(
                                                                ex.filename))
                    retval = Prog.RetVal.exit_failure
                    continue
                else:
                    try:
                        target_file = pathlib.Path(os.readlink(str(dane_l)))
                    except OSError as ex:
                        prog.log.error(
                                "dane file '{}' is not a symlink".format(
                                                                  ex.filename))
                        retval = Prog.RetVal.exit_failure
                        continue

                    #if not target_file.is_symlink():
                    #    prog.log.error("dane symlink '{}' resolves to '{}', which is not a symlink itself".format(dane_l, target_file))
                    #    retval = False
                    #    continue
            except OSError as ex:
                prog.log.error("recreating symlink '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
                retval = Prog.RetVal.exit_failure
                continue

    if not retval == Prog.RetVal.ok:
        return retval

    # the dane directory and symlinks have been created (or are fine). Now,
    # every domain in the prog.target_list MUST exist in
    # prog.dane_domain_directories:
    for t in prog.target_list:
        if t.domain not in prog.dane_domain_directories:
            prog.log.error("domain '{}' from config file does not have a corresponding letsencrypt directory".format(t.domain))
            retval = Prog.RetVal.exit_failure

    # now, prog.dane_domain_directories is a dictionary containing all
    # directory names in the letsencrypt live directory, but we're not
    # interested in _all_ of them, just the ones being used by this
    # program; that is, only those defined in the config file (and hence
    # in prog.target_list). So, let's remove the other ones:
    new_dict = { }
    for d in prog.dane_domain_directories:
        if d in [ t.domain for t in prog.target_list ]:
            new_dict[d] = prog.dane_domain_directories[d]

    prog.dane_domain_directories = new_dict
    prog.log.info3(" ++ dane_domain_directories: {}".format(
                                                prog.dane_domain_directories))

    return retval


def set_renewed_domains(prog):
    """Check if the environment parameter 'RENEWED_DOMAINS' is set.

    Args:
        prog (State): program internal state.

    Returns:
        RetVal: always returns 'RetVal.ok'.
    """
    prog.log.info1("+++ checking environment parameter 'RENEWED_DOMAINS'")
    try:
        prog.renewed_domains = os.environ['RENEWED_DOMAINS'].split()
    except KeyError:
        # Maybe never return an error if not set, even in posthook mode.
        #if prog.args.post:
        #    prog.log.error(
        #            "no RENEWED_DOMAINS parameter set in the environment")
        #    return Prog.RetVal.exit_failure
        prog.log.info1("  + not set")
        return Prog.RetVal.ok

    for i in prog.renewed_domains:
        prog.log.info1("  + {}".format(i))
    if not prog.renewed_domains:
        prog.log.info1("  + empty")

    return Prog.RetVal.ok


def live_to_archive(prog):
    """Move dane symlinks from pointing to live certs to archive certs.

    Args:
        prog (State): program internal state.

    Returns:
        RetVal: returns 'RetVal.continue_failure' if any error is
            encountred, and returns 'RetVal.ok' otherwise.
    """
    prog.log.info1("+++ moving dane symlinks to point from live to archive")
    retval = Prog.RetVal.ok

    # loop over target_list rather than the keys in dane_domain_directories
    # since we'll be adding certs to the target
    for t in prog.target_list:
        for l in prog.dane_domain_directories[t.domain]:

            # the (full path) dane symlink
            dane_l = prog.dane_directory / t.domain / l
            prog.log.info2(" ++ dane: {}".format(dane_l))

            try:
                # the dane file MUST resolve to something (i.e. be a symlink)
                resolv1 = pathlib.Path(os.readlink(str(dane_l)))
            except OSError as ex:
                prog.log.error(
                        "dane file '{}' is not a symlink".format(ex.filename))
                retval = Prog.RetVal.continue_failure
                continue

            prog.log.info3("    => {}".format(resolv1))

            if not resolv1.is_absolute():
                resolv1 = dane_l.parent / resolv1

                try:
                    presolv1 = resolv1.parent.resolve()
                except FileNotFoundError as ex:
                    prog.log.error(
                       "path in live certificate file '{}' not found".format(
                                                               ex.filename))
                    retval = Prog.RetVal.continue_failure
                    continue
                except RuntimeError as ex:
                    prog.log.error("recursive loop in resolving live certificate file '{}'".format(ex.filename))
                    retval = Prog.RetVal.continue_failure
                    continue

                resolv1 = presolv1 / resolv1.name


            # 'resolv1' may or may not be a symlink: ordinarily it should
            # be a symlink, but if it's already been processed by a
            # pre-hook command, then it will be an archive (regular) file.
            if resolv1.is_symlink():
                try:
                    archive_f = pathlib.Path(os.readlink(str(resolv1)))
                except OSError as ex:
                    prog.log.error(
                        "live file '{}' is not a symlink".format(ex.filename))
                    retval = Prog.RetVal.continue_failure
                    continue

                prog.log.info3("       => {}".format(archive_f))

            else:
                # if not a symlink, we assume it's been processed by
                # prehook already.
                prog.log.info2("    points to a regular file")
                continue

            if not archive_f.is_absolute():
                archive_f = resolv1.parent / archive_f
                try:
                    archive_f = archive_f.resolve()
                except FileNotFoundError as ex:
                    prog.log.error("path in archive certificate file '{}' not found".format(ex.filename))
                    retval = Prog.RetVal.continue_failure
                    continue
                except RuntimeError as ex:
                    prog.log.error("recursive loop in resolving archive certificate file '{}'".format(ex.filename))
                    retval = Prog.RetVal.continue_failure
                    continue

            try:
                dane_l.unlink()
            except FileNotFoundError as ex:
                prog.log.error(
                        "dane symlink '{}' not found".format(ex.filename))
                retval = Prog.RetVal.continue_failure
                continue
            except OSError as ex:
                prog.log.error(
                        "removing dane symlink '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
                retval = Prog.RetVal.continue_failure
                continue

            try:
                dane_l.symlink_to( relative_to(dane_l, archive_f) )
            except FileExistsError as ex:
                # seeing as we deleted this file soon before this point...
                prog.log.error(
                    "recreating symlink '{}': file exists".format(ex.filename))
                retval = Prog.RetVal.continue_failure
                continue
            except OSError as ex:
                prog.log.error("recreating symlink '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
                retval = Prog.RetVal.exit_failure
                continue

            prog.log.info2("    => moved to {}".format(archive_f))

            t.add_cert(dane_l, resolv1, archive_f)

    return retval


def process_data(prog):
    """Process datafile lines.

    Data is organized in 'groups', which are collections of datafile lines
    that share a common domain. First, withing every group, all the
    delete lines are processed. Then we loop over the groups again, and if
    there exist posthook lines, we call 'process_data_posthook', otherwise
    we call 'process_data_prehook'.

    Args:
        prog (State): program internal state.

    Returns:
        RetVal: returns 'RetVal.continue_failure' if any error is
            encountred, and returns 'RetVal.ok' otherwise.
    """
    retval = Prog.RetVal.ok

    for group in prog.data.groups:
        # if there are any delete lines, we should try to process them now
        for l in group.special:
            try:
                delete_dane_if_up(prog, group.target.api, l.tlsa, l.hash)
                l.write_state_off()
            except Except.DNSSkip as ex:
                prog.log.info2("  + {}".format(ex.message))
                prog.log.info2(
                        "  + TLSA record not removed; incrementing the count")
                l.increment_count()
            except Except.DNSNoReturnError as ex:
                prog.log.error(ex.message)
                prog.log.info2(
                        "  + TLSA record not removed; incrementing the count")
                l.increment_count()
            except (Except.DNSError, Except.InternalError) as ex:
                prog.log.error(ex.message)
                prog.log.info2(
                        "  + TLSA record not removed; incrementing the count")
                l.increment_count()
                retval = Prog.RetVal.continue_failure

    for group in prog.data.groups:
        if group.post:
            if process_data_posthook(prog, group):
                retval = Prog.RetVal.continue_failure
        else:
            if process_data_prehook(prog, group):
                retval = Prog.RetVal.continue_failure

    return retval


def exit(prog, value=0):
    """Exit depending on the internal program state."""
    # so far, the only error setting that might be set is prog.log.errors.
    if prog.log.has_errors():
        value += 16
    prog.log.info3("+++ exiting with code: {}".format(value))
    sys.exit(value)



def delete_dane_if_up(prog, api, tlsa, hash1, hash2 = None):
    """Delete a DANE TLSA record.

    Basically, all this function does is attempt to delete 'hash1':

    hash1 = {the old TLSA hash: need to see if it is up and then delete it}
    hash2 = None

    hash1 = {the old TLSA hash: if it AND hash2 are up, then delete it}
    hash2 = {the new TLSA hash: already published, needs to be up before hash1
             can be deleted}

    This function is called in three situations:
        1. when reading a delete line, so basically when told to delete a
            record.
        2. when reading a posthook line on a non-renewed domain, after the TTL
            value has passed and we want to delete an old hash
        3. when reading a posthook line on a renewed domain whose pending state
            is '0', so basically we have renewed again and we need to delete
            the previous 'new' hash that was published.

    Note that a failure of 1. causes repeated attempts at 1; a failure of 2
    causes repeated attempts at 2; and a failure of 3 causes a delete line to
    be made, and so causes 1.

    Args:
        prog (State): program internal state.

    Returns:
        NoneType: always returns 'None'.

    Raises:
        DNSNotLive: if DANE record not up yet.
        DNSProcessingError: if an error occurs that should cause Alnitak
            to exit with an error code.
        DNSNoReturnError: if an error occurs that requests that Alnitak do
            not exit with an error code (raised from 'binary.delete').
    """
    prog.log.info1(
            "+++ attempting to delete TLSA DNS record: {}".format(tlsa.pstr()))
    if api.type == Prog.ApiType.cloudflare4:

        # get a dict of all the records up
        records = cloudflare4.read(prog, api, tlsa)

        # if we need to check if hash2 is up, then do that now
        if hash2:
            for r in records:
                if r == hash2:
                    break
            else:
                raise Except.DNSNotLive("TLSA record not up yet")

        # if the hash is in 'records', then we need to delete that
        # record
        for r in records:
            if r == hash1:
                cloudflare4.delete(prog, api, tlsa, records[r])
                break
        else:
            raise Except.DNSNotLive("TLSA record not up yet")

    else:
        binary.delete(prog, api, tlsa, hash1, hash2)


def process_data_prehook(prog, group):
    """Process prehook lines.

    If the domain in the group is renewed, then call 'publish_dane',
    otherwise call 'archive_to_live'.

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    if group.domain in prog.renewed_domains:
        prog.log.info1(
                "+++ prehook line: domain '{}' renewed".format(group.domain))
        if group.target:
            return publish_dane(prog, group)
        else:
            prog.log.error("domain '{}' renewed, but not in config file: cannot process!".format(group.domain))
            return True
    else:
        prog.log.info1(
            "+++ prehook line: domain '{}' not renewed".format(group.domain))
        # move dane symlinks back to live
        return archive_to_live(prog, group)


def process_data_posthook(prog, group):
    """Process posthook lines.

    If the domain in the group is renewed, then call
    'process_data_posthook_renewed', otherwise call
    'process_data_posthook_not_renewed'.

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    if group.domain in prog.renewed_domains:
        prog.log.info1(
                "+++ posthook line: domain '{}' renewed".format(group.domain))
        return process_data_posthook_renewed(prog, group)

    else:
        prog.log.info1(
            "+++ posthook line: domain '{}' not renewed".format(group.domain))
        # check the posthook line state; if '0', then see if the dane is up
        # and then delete the line if it is; if '1', then retry to publish
        # the dane
        return process_data_posthook_not_renewed(prog, group)


def process_data_posthook_not_renewed(prog, group):
    """Process posthook lines for the non-renewed domain.

    For all lines in the group, if the line's pending state is '0', then
    delete the old TLSA record if the time-to-live value has passed,
    otherwise, try to publish the TLSA record again (since it must have
    failed last time).

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    errors = False

    for l in group.post:
        if l.pending == '0':
            prog.log.info1(" ++ pending state is 0: ({}): deleting old TLSA DNS record".format(l.tlsa.pstr()))
            try:
                # check if TTL has passed
                prog.log.info2(" ++ checking TTL value has passed")
                #check_ttl(l.data.time)
                try:
                    time_now = int("{:%s}".format(prog.timenow))
                    time_published = int(l.time)
                except ValueError as ex:
                    raise Except.InternalError("time value is not an integer")

                time_passed = time_now - time_published
                prog.log.info3("  + published at: {}\n  + now: {} ({} seconds elapsed)".format(time_published, time_now, time_passed))

                if time_passed < prog.ttl:
                    raise Except.DNSSkipProcessing("time to live value ({}) hasn't passed: {} seconds remain".format(prog.ttl, prog.ttl - time_passed))


                # get cert hash
                cert = certop.get_archive(l.tlsa, group.pre)
                prog.log.info2(
                        "  + old hash: going to use cert '{}'".format(cert))

                cert_data = certop.read_cert(cert, l.tlsa)
                hash = certop.get_hash( l.tlsa.selector, l.tlsa.matching,
                                             cert_data)
                prog.log.info2("  + old {}{}{} hash: {}".format(
                    l.tlsa.usage, l.tlsa.selector, l.tlsa.matching, hash))

                # check if the dns record is up
                delete_dane_if_up(prog, group.target.api, l.tlsa, hash, l.hash)

                # change the write state of the line
                l.write_state_off()

            except Except.DNSSkip as ex:
                prog.log.info2("  + {}".format(ex.message))
            except (Except.DNSError, Except.InternalError) as ex:
                prog.log.error(ex.message)
                errors = True
        else:
            # retry publishing record
            prog.log.info1(" ++ pending state is 1: will retry publishing TLSA DNS record {}".format(l.tlsa.pstr()))
            try:
                if group.target.api.type == Prog.ApiType.cloudflare4:
                    cloudflare4.publish(prog, group.target.api, l.tlsa, l.hash)
                else:
                    binary.publish(prog, group.target.api, l.tlsa, l.hash)

                prog.log.info2("  + record published successfully")

                # switch the pending state of the line
                l.pending_off()

                # update the time
                l.change_time("{:%s}".format(prog.timenow))

            except Except.DNSSkip as ex:
                prog.log.info2("  + {}".format(ex.message))
                # If the record is already up, then we do not need to do any
                # further processing
                l.write_state_off()
            except Except.DNSNoReturnError as ex:
                prog.log.error(ex.message)
            except (Except.DNSError, Except.InternalError) as ex:
                prog.log.error(ex.message)
                errors = True

    # let's now see if we have _any_ posthook lines to write
    for l in group.post:
        if l.state == Prog.DataLineState.write:
            break
    else:
        # no posthook lines: remove all the prehook lines:
        prog.log.info2(
                "+++ all posthook lines processed (all records up/deleted)")
        if archive_to_live(prog, group):
            errors = True

    return errors


def process_data_posthook_renewed(prog, group):
    """Process posthook lines for the renewed domain.

    For all lines in the group, if the line's pending state is '0', then
    mark the old 'new' TLSA record for deletion, publish the renewed
    certificate's TLSA record and then try deleting the old record. If the
    pending state is '1' then just publish the renewed certificate's
    TLSA record.

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    # if the domain is renewed, for every posthook line, we check the pending
    # state: if it is '0', then we try to _delete_ the recorded hash if it is
    # up. Q: what happens if _this_ fails? A: we create a new special DEL
    # line. We then switch the state of the line to skip and then switch the
    # pending state of all the prehook lines. We then run the
    # 'process_data_prehook' function.

    errors = False
    prog.log.info1(" ++ certificate renewed again!")
    for l in group.post:
        if l.pending == '0':

            prog.log.info2(
                    " ++ will check if new cert hash matches the previous one")
            # if, for some reason, the re-renewed certificate has the same
            # hash as the previously renewed certificate, we neither want to
            # bother deleting the hash nor bother publishing that hash again
            # in 'process_data_prehook'.
            cert = certop.get_live(l.tlsa, group.pre)
            prog.log.info2("  + going to use cert '{}'".format(cert))

            cert_data = certop.read_cert(cert, l.tlsa)

            hash = certop.get_hash(l.tlsa.selector, l.tlsa.matching, cert_data)

            prog.log.info2("  + {}{}{} hash: {}".format(
                        l.tlsa.usage, l.tlsa.selector, l.tlsa.matching, hash))

            if hash == l.hash:
                # what happens here is this: we have a posthook line that
                # tells us that a previous 'new' TLSA record was published
                # (pending state is '0'). The new live certificate has the
                # same hash. What we want to do is keep the line as it is
                # (do not update the time) since this new certificate has
                # effectively already been published. Normally, if this were
                # not the case we would delete the record (and create a delete
                # line if this failed), switch the line state and then act as
                # if the previous renewal never happened by switching the
                # prehook line pending states to '0'. We would then call
                # 'process_data_prehook' to publish new TLSA records and write
                # new posthook lines. Since we will still do this (since there
                # might be TLSA records that _have_ changed), we need to
                # signify that this particular TLSA record that did not change
                # does not need to be processed. So, we switch the tlsa state
                # (to False) and leave the posthook line as it is.
                prog.log.info2("  + hash has not changed; record is already up: update posthook line")
                # 'process_data_prehook', which will be called at the end of
                # this function, uses the tlsa records in group.target, and
                # not the tlsa object in the DataPost objects, so let's find
                # the matching tlsa record and change the state of that:
                for t in group.target.tlsa:
                    if l.tlsa == t:
                        t.publish_off()
                        break
                continue


            prog.log.info2(" ++ previous 'new' TLSA DNS record ({}) needs to be deleted: will do so after the new records are published...".format(l.tlsa.pstr()))
            l.mark_for_deletion()

        # no matter what the pending state ('0' if the old 'new' certs
        # published a record or '1' if the old 'new' certs did not publish a
        # record) we always want to delete this line (do not write this line)
        # since the is the line for the old new cert and the certs have been
        # renewed again, EXCEPT whenthe new new cert matches the old new
        # cert, in which case we want to keep this line as being written,
        # which we do by continuing past this next line if so:
        l.write_state_off()
        #l.change_time("{:%s}".format(prog.timenow))

    for l in group.pre:
        l.pending_off()

    errors = process_data_prehook(prog, group)

    prog.log.info2("+++ deleting records previously marked for deletion...")
    for l in group.post:
        if l.mark_delete:
            try:
                delete_dane_if_up(prog, group.target.api, l.tlsa, l.hash)
            except Except.DNSSkip as ex:
                prog.log.info2("  + {}".format(ex.message))
                prog.log.info3("  + will write a delete line")
                group.add_special(
                    Prog.DataDelete(
                                group.domain, 0, l.tlsa, '1', l.time, l.hash) )
            except Except.DNSNoReturnError as ex:
                prog.log.error(ex.message)
                prog.log.info3("  + will write a delete line")
                group.add_special(
                    Prog.DataDelete(
                                group.domain, 0, l.tlsa, '1', l.time, l.hash) )
            except (Except.DNSError, Except.InternalError) as ex:
                prog.log.error(ex.message)
                prog.log.info3("  + will write a delete line")
                group.add_special(
                    Prog.DataDelete(
                                group.domain, 0, l.tlsa, '1', l.time, l.hash) )
                errors = True

    return errors


def publish_dane(prog, group):
    """Publish a DANE record.

    Called in posthook mode, for renewed or non-renewed domains, this
    function will try to publish DANE TLSA records.

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    errors = False

    for tlsa in group.target.tlsa:
        if not tlsa.publish:
            prog.log.info1(" ++ TLSA record already up: {}".format(tlsa.pstr()))
            continue

        prog.log.info1(
                " ++ will attempt to publish TLSA DNS record: {}".format(
                                                                  tlsa.pstr()))
        pending = '0'
        try:
            cert = certop.get_live(tlsa, group.pre)
            prog.log.info2("  + going to use cert '{}'".format(cert))

            cert_input = certop.read_cert(cert, tlsa)

            hash = certop.get_hash(tlsa.selector, tlsa.matching,
                                        cert_input)

            prog.log.info2("  + {}{}{} hash: {}".format(
                            tlsa.usage, tlsa.selector, tlsa.matching, hash))

            # now need to use the Api object to publish a TLSA record
            if group.target.api.type == Prog.ApiType.cloudflare4:
                cloudflare4.publish(prog, group.target.api, tlsa, hash)
            else:
                binary.publish(prog, group.target.api, tlsa, hash)

        except Except.DNSSkip as ex:
            # e.g. this is likely to happen for DANE-TA(2) records, whose
            # hashes do not change as frequently. In this case we do not add
            # any posthook line
            prog.log.info2("  + record is already up; will treat all further processing as completed")
            continue
        except Except.DNSNoReturnError as ex:
            prog.log.error(ex.message)
            pending = '1'
        except (Except.DNSError, Except.InternalError) as ex:
            prog.log.error(ex.message)
            pending = '1'
            errors = True

        prog.log.info3(
                "  + creating posthook line with pending '{}'".format(pending))
        group.add_post( Prog.DataPost( group.domain, 0, tlsa, pending,
                                  "{:%s}".format(prog.timenow), hash) )

    if group.post:
        for l in group.pre:
            l.pending_on()
    else:
        errors2 = archive_to_live(prog, group)
        return errors or errors2

    return errors


def archive_to_live(prog, group):
    """Move dane symlinks from pointing to archive certs to back to live certs.

    Args:
        prog (State): program internal state.
        group (DataGroup): the group of prehook and/or posthook lines.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    prog.log.info1(" ++ moving dane symlinks to point from archive to live")
    errors = False

    for l in group.pre:
        prog.log.info3("  + dane: {}".format(l.cert.dane))
        if not create_symlink(prog, l.cert.dane, l.cert.live):
            errors = True
            continue

        # change the state of the line
        l.write_state_off()
        prog.log.info3("    => {} (state: {})".format(l.cert.live, l.state))

    return errors

# TODO: use me? use me!
def create_symlink(prog, symlink, to):
    """Create a symlink.

    Note that this function will try to write relative-path symlinks.

    Args:
        prog (State): program internal state.
        symlink (pathlib.Path): symlink file to create.
        to (pathilb.Path): path the symlink file points to. Can be an
            absolute or relative path: this function will handle either
            case properly.

    Returns:
        bool: return 'True' for errors, 'False' otherwise.
    """
    try:
        symlink.unlink()
    except FileNotFoundError as ex:
        pass
    except OSError as ex:
        prog.log.error("removing symlink file '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return False

    try:
        symlink.symlink_to( relative_to(symlink, to) )
    except FileExistsError as ex:
        # seeing as we deleted this file soon before this point...
        prog.log.error(
                "recreating symlink '{}': file exists".format(ex.filename))
        return False
    except OSError as ex:
        prog.log.error("recreating symlink '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return False

    return True


def relative_to(path, target):
    """Return the relative path of the input to the target path.

    For example:
        path = /a/b/d/c/Y
        target = /a/b/c/X
    Then this function will return:
        '../../c/X'

    Args:
        path (pathlib.Path):
        target (pathlib.Path):

    Returns:
        str: the path of 'path' relative to 'target'.
    """
    #   we step back through 'path' until the parent directory matches a
    #   parent directory of 'target'. We then construct the return by
    #   adding N (../) paths for the N steps back and then appending the
    #   path in 'target' that was not matched in the comparison.
    for n, p in enumerate(path.parents):
        try:
            q = target.relative_to(p)
        except ValueError:
            continue
        return "../" * n + str(q)
    return str(target)








def main():
    """Alnitak program entry function."""

    prog = Prog.State()

    parser = argparse.ArgumentParser(
            prog=prog.name,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [options]",
            description="""This program helps to automatically manage DANE records when Let's Encrypt
certificates are renewed.""",
            epilog="""{} {}
{}""".format(prog.name, prog.version, prog.copyright)
            )

    parser.add_argument("-V", "--version",
            help="print the program version and exit.",
            action='store_true')

    mode = parser.add_mutually_exclusive_group()

    mode.add_argument("-t", "--config-test",
            help="test the config file for errors",
            action='store_true')

    mode.add_argument("--pre",
            help="run in pre-hook mode, to be run on the Let's Encrypt pre-hook function",
            action='store_true')

    mode.add_argument("--post",
            help="run in post-hook mode, to be run on the Let's Encrypt post-hook function",
            action='store_true')

    mode.add_argument("--reset",
            help="recreate the symlinks in the dane directory",
            action='store_true')

    parser.add_argument("--dane-directory",
            help="directory containing the DANE certificates")

    parser.add_argument("--letsencrypt-directory",
            help="directory containing the Let's Encrypt certificates")

    parser.add_argument("--ttl",
            help="time to live (seconds)",
            type=int)

    parser.add_argument("-l", "--log",
            help="log to destination file")

    parser.add_argument("-L", "--log-level",
            help="set the logging level",
            choices=["normal", "verbose", "full", "no"])

    parser.add_argument("-q", "--quiet",
            help="suppress output to stdout and stderr (but not logfile)",
            action='store_true')

    parser.add_argument("-c", "--config",
            help="read the specified config file")


    args = parser.parse_args()
    exec_list = None
    errors = False

    if args.version:
        print("{} {}\n{}".format(prog.name, prog.version, prog.copyright))
        sys.exit(0)

    if args.ttl:
        if args.ttl < prog.ttl_min:
            print("{}: ttl value must be at least '{}'.".format(
                                                     prog.name, prog.ttl_min))
            sys.exit(2)
        prog.set_ttl(args.ttl)

    # must set 'prog.log.quiet' before the other 'prog.log' details.
    if args.quiet:
        prog.log.quiet = True

    if args.log:
        if args.log == 'stdout' or args.log == '-':
            prog.log.set_stdout()
        elif args.log == 'no':
            prog.log.set_nolog()
        else:
            if args.log[-1] == '/':
                prog.log.set_file("{}/{}.log".format(args.log, prog.name))
            else:
                prog.log.set_file(args.log)

    if args.log_level:
        if args.log_level == 'verbose':
            prog.log.set_level(Prog.LogLevel.verbose)
        elif args.log_level == 'full':
            prog.log.set_level(Prog.LogLevel.full)
        elif args.log_level == 'no':
            prog.log.set_level(Prog.LogLevel.nolog)

    if args.dane_directory:
        prog.set_dane_directory(args.dane_directory)

    if args.letsencrypt_directory:
        prog.set_letsencrypt_directory(args.letsencrypt_directory)

    if args.config:
        prog.set_config_file(args.config)


    if args.config_test:
        exec_list = [ config.read ]

    if args.reset:
        prog.recreate_dane = True
        exec_list = [ config.read, init_dane_directory, datafile.remove ]

    if args.pre:
        exec_list = [ config.read, init_dane_directory, live_to_archive,
                      datafile.write_prehook ]

    if args.post:
        exec_list = [ config.read, set_renewed_domains, datafile.read,
                      datafile.check_data, process_data,
                      datafile.write_posthook ]

    if not exec_list:
        exec_list = [ config.read, set_renewed_domains, datafile.read,
                      datafile.check_data, process_data,
                      datafile.write_posthook ]


    # first create a lock:
    try:
        if prog.lock():
            if not args.quiet:
                print("alnitak: another instance is already running.")
            sys.exit(32)
                # don't call exit(prog, 32): nothing worth logging, so if
                # there's an error, we'll just ignore it.
    except Except.LockError as ex:
        if not args.quiet:
            print("error: {}.".format(ex.message), file=sys.stderr)
        sys.exit(4)
            # don't call exit(prog, 4): nothing worth logging, so if
            # there's an error, we'll just ignore it.


    # next initialize logging
    prog.init_logging(args)


    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                       ENTRY FUNCTIONS                             #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    #            PRE                                POST
    #    [1]  read_config                   [1]  read_config
    #    [2]  init_dane_directory           [3]  set_renewed_domains
    #    [6]  live_to_archive               [4]  read_datafile
    #    [8]  write_prehook_datafile        [5]  check_data
    #                                       [7]  process_data
    #           RESET                       [9]  write_posthook_datafile
    #    [1]  read_config
    #    [2]  init_dane_directory                   TEST
    #    [10] remove_datafile               [1]  read_config


    # then run the rest of the code
    for prog_call in exec_list:
        retval = prog_call(prog)
        if retval == Prog.RetVal.ok:
            continue
        elif retval == Prog.RetVal.exit_ok:
            exit(prog)
        elif retval == Prog.RetVal.continue_failure:
            errors = True
        else:
            exit(prog, retval.value)

    if errors:
        exit(prog, 1)
    exit(prog)


if __name__ == "__main__":
    main()





import requests

from alnitak import exceptions as Except


# For reference, the status codes are responses are as follows:
#
### Key is wrong (not enough characters):
#   400
#   {'messages': [],
#    'success': False,
#    'errors': [
#           {'error_chain': [
#                    {'message': 'Invalid format for X-Auth-Key header',
#                     'code': 6103}
#                ],
#            'message': 'Invalid request headers',
#            'code': 6003}
#        ],
#    'result': None}
#
### Key is wrong (invalid):
#   403
#   {'messages': [],
#    'success': False,
#    'errors': [
#           {'message': 'Unknown X-Auth-Key or X-Auth-Email',
#            'code': 9103}
#        ],
#    'result': None}
#
### Record is already up:
#   400
#   {'messages': [],
#    'success': False,
#    'errors': [
#           {'message': 'The record already exists.',
#            'code': 81057}
#       ],
#    'result': None}
#
#


def get_errors(response):
    """Extract error messages from the JSON response.

    Argument should be a dict. If a key called 'errors' is found, it
    should be a list:
        { 'errors': [ ... ], ... }

    The objects in the list should be a dict. If this dict has keys
    'message' and 'code', then their values will be added to a list
    of error messages to return:
        { 'message': 'X', 'code': N }  -->  [ [N, 'X'] ]

    If the dict has a key 'error_chain', then its value should be a
    list of dicts also containing 'message' and 'code' as keys; these
    will also be added to the list of error messages to return:
        { 'message': 'X',
          'code': N,
          'error_chain': [ {'message': 'Y', 'code': P} ] }
    will construct:
        [ [P, 'Y'], [N, 'X'] ]

    The complete error list to return will be constructed from all
    such list items of 'errors' found in the argument dict in this
    way.

    Args:
        response (dict): JSON response from a 'requests' call.

    Returns:
        list(list(int, str)): return list a lists containing an (int)
            error code and (str) message.
    """
    errors = []
    for a in response['errors']:
        try:
            temp = [ [ b['code'], b['message'] ] for b in a['error_chain'] ]
        except KeyError:
            temp = []

        errors += temp

        try:
            errors += [ [a['code'], a['message']] ]
        except KeyError:
            pass

    return errors


def delete(prog, api, tlsa, id):
    """Delete a DANE TLSA record.

    Args:
        prog (State): not changed.
        api (ApiCloudflare4): contains Cloudflare4 login details.
        tlsa (Tlsa): details of the DANE TLSA record to delete.
        id (str): The Cloudflare ID of the record to delete.

    Returns:
        NoneType: always returns 'None'.

    Raises:
        DNSProcessingError: if an error ocurred at any point that should
            cause the Alnitak to exit with an error exit code.
    """
    prog.log.info2("  + deleting TLSA record for {}".format(tlsa.pstr()))

    headers = { "X-Auth-Email": api.email,
                "X-Auth-Key": api.key,
                "Content-Type": "application/json" }

    try:
        r = requests.delete("https://api.cloudflare.com/client/v4/zones/{}/dns_records/{}".format(api.zone, id), headers=headers)
    except ConnectionError:
        raise Except.DNSProcessingError("connection error encountered")
    except requests.exceptions.Timeout:
        raise Except.DNSProcessingError("request timed out")
    except requests.exceptions.TooManyRedirects:
        raise Except.DNSProcessingError("too many redirects")
    except requests.exceptions.RequestException as ex:
        raise Except.DNSProcessingError("{}".format(ex))

    prog.log.info3("  + HTTP response: {}".format(r.status_code))

    response = r.json()
    prog.log.info3("  + JSON response: {}".format(response))

    errors = get_errors(response)
    if errors:
        raise Except.DNSProcessingError(errors)

    if r.status_code >= 400 and r.status_code <= 599:
        raise Except.DNSProcessingError(
                "Cloudflare4 HTTP response was {}".format(r.status_code))

    if not response['success']:
        raise Except.DNSProcessingError("Cloudflare4 JSON response failure")



def publish(prog, api, tlsa, hash):
    """Create (publish) a DANE TLSA record.

    Args:
        prog (State): not changed.
        api (ApiCloudflare4): contains Cloudflare4 login details.
        tlsa (Tlsa): details of the DANE TLSA record to delete.
        hash (str): DANE TLSA 'certificate data' (hash) to publish.

    Returns:
        NoneType: always returns 'None'.

    Raises:
        DNSProcessingError: if an error ocurred at any point that should
            cause the Alnitak to exit with an error exit code.
        DNSSkipProcessing: if the record is already up.
    """
    prog.log.info2("  + publishing TLSA record for {}".format(tlsa.pstr()))

    headers = { "X-Auth-Email": api.email,
                "X-Auth-Key": api.key,
                "Content-Type": "application/json" }

    data = '{{ "type": "TLSA", "name": "_{}._{}.{}", "data": {{ "usage": {}, "selector": {}, "matching_type": {}, "certificate": "{}" }} }}'.format(
                    tlsa.port, tlsa.protocol, tlsa.domain, tlsa.usage,
                    tlsa.selector, tlsa.matching, hash)

    try:
        r = requests.post("https://api.cloudflare.com/client/v4/zones/{}/dns_records".format(api.zone), data=data, headers=headers)
    except ConnectionError:
        raise Except.DNSProcessingError("connection error encountered")
    except requests.exceptions.Timeout:
        raise Except.DNSProcessingError("request timed out")
    except requests.exceptions.TooManyRedirects:
        raise Except.DNSProcessingError("too many redirects")
    except requests.exceptions.RequestException as ex:
        raise Except.DNSProcessingError("{}".format(ex))

    prog.log.info3("  + HTTP response: {}".format(r.status_code))

    response = r.json()
    prog.log.info3("  + JSON response: {}".format(response))

    errors = get_errors(response)

    # record is already up
    if len(errors) == 1 and errors[0][0] == 81057:
        # we will only accept this code if it is the only error
        # encountered
        raise Except.DNSSkipProcessing(errors[0][1])

    if errors:
        raise Except.DNSProcessingError(errors)

    if r.status_code >= 400 and r.status_code <= 599:
        raise Except.DNSProcessingError(
                "Cloudflare4 HTTP response was {}".format(r.status_code))

    if not response['success']:
        raise Except.DNSProcessingError("Cloudflare4 JSON response failure")



def read(prog, api, tlsa):
    """Get a dict of DANE TLSA records that are up.

    For example, if the following TLSA records are up:
        TLSA 3 1 1 _25._tcp.example.com  abcABC
        TLSA 3 1 1 _25._tcp.example.come defDEF
    and 'tlsa' has param '311', port '25', protocol 'tcp' and domain
    'example.com', then this function will return:
        { "id...": "abcABC", "id...": "defDEF" }
    where "id..." will be some unique ID Cloudflare has assigned to that
    record.
 
    Args:
        prog (State): not changed.
        api (ApiCloudflare4): contains Cloudflare4 login details.
        tlsa (Tlsa): details of the DANE TLSA record to retrieve.

    Returns:
        NoneType: always returns 'None'.

    Raises:
        DNSProcessingError: if an error ocurred at any point that should
            cause the Alnitak to exit with an error exit code.
        DNSNotLive: if no matching records are up.
    """
    prog.log.info2("  + getting TLSA records for _{}._{}.{}".format(
                                        tlsa.port, tlsa.protocol, tlsa.domain))

    headers = { "X-Auth-Email": api.email,
                "X-Auth-Key": api.key,
                "Content-Type": "application/json" }

    params = { "type": "TLSA",
               "name":"_{}._{}.{}".format(
                                      tlsa.port, tlsa.protocol, tlsa.domain) }

    try:
        r = requests.get("https://api.cloudflare.com/client/v4/zones/{}/dns_records".format(api.zone), params=params, headers=headers)
    except ConnectionError:
        raise Except.DNSProcessingError("connection error encountered")
    except requests.exceptions.Timeout:
        raise Except.DNSProcessingError("request timed out")
    except requests.exceptions.TooManyRedirects:
        raise Except.DNSProcessingError("too many redirects")
    except requests.exceptions.RequestException as ex:
        raise Except.DNSProcessingError("{}".format(ex))

    prog.log.info3("  + HTTP response: {}".format(r.status_code))

    response = r.json()
    prog.log.info3("  + JSON response: {}".format(response))

    errors = get_errors(response)
    if errors:
        raise Except.DNSProcessingError(errors)

    if r.status_code >= 400 and r.status_code <= 599:
        raise Except.DNSProcessingError(
                "Cloudflare4 HTTP response was {}".format(r.status_code))

    if not response['success']:
        raise Except.DNSProcessingError("Cloudflare4 JSON response failure")

    ret = {}
    for r in response['result']:
        ret[r['data']['certificate'].lower()] = r['id']

    if ret:
        return ret

    raise Except.DNSNotLive("no TLSA records found")






import os
import re

from alnitak import prog as Prog


def read(prog):
    """Read a datafile and set data in the internal program state.

    Args:
        prog (State): data is set based on the contents of the datafile.
            No datafile is not taken as an error: simply nothing is done.

    Returns:
        RetVal: returns 'RetVal.ok' if no errors encountered,
            'RetVal.exit_ok' if no datafile found, 'RetVal.exit_failure'
            if errors found in the datafile or opening/reading the datafile
            failed.
    """
    prog.log.info1("+++ reading datafile '{}'".format(prog.datafile))
    retval = Prog.RetVal.ok

    try:
        with open(str(prog.datafile), "r") as file:
            raw = file.read().splitlines()
    except FileNotFoundError as ex:
        # if there is no datafile, then posthook has nothing to do: we should
        # just exit.
        prog.log.info1("  + no file to read")
        return Prog.RetVal.exit_ok
    except OSError as ex:
        prog.log.error(
                "datafile '{}': {}".format(ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure


    line_pos = 0

    for l in raw:
        line_pos += 1

        # matches empty line
        match = re.match(r'^\s*(#.*)?$', l)
        if match:
            continue

        # prehook line
        #   domain  dane_cert  live_cert  archive_cert  pending
        match = re.match(r'(?P<domain>{})\s+"(?P<dane>(\\.|[^"])+)"\s+"(?P<live>(\\.|[^"])+)"\s+"(?P<archive>(\\.|[^"])+)"\s+(?P<pending>(0|1))'.format(prog.tlsa_domain_regex), l)
        if match:

            prog.log.info3("  + line {}: prehook line (pending: {})".format(
                                            line_pos, match.group('pending')))

            prog.data.add_line( prog, Prog.DataPre( match.group('domain'),
                                               line_pos,
                                               match.group('dane'),
                                               match.group('live'),
                                               match.group('archive'),
                                               match.group('pending') ) )

            continue

        # posthook line
        #   x.com 301 25 tcp x.com unix_time 0 hash
        match = re.match(r'(?P<domain>{})\s+(?P<tlsa_spec>{})\s+(?P<tlsa_port>[0-9]+)\s+(?P<tlsa_protocol>{})\s+(?P<tlsa_domain>{})\s+(?P<time>[0-9]+)\s+(?P<pending>(0|1))\s+(?P<hash>[a-fA-F0-9]+)'.format(prog.tlsa_domain_regex, prog.tlsa_parameters_regex, prog.tlsa_protocol_regex, prog.tlsa_domain_regex), l)
        if match:

            prog.log.info3("  + line {}: posthook line (pending: {})".format(
                                            line_pos, match.group('pending')))

            prog.data.add_line( prog, Prog.DataPost( match.group('domain'),
                                                line_pos,
                                                Prog.Tlsa(
                                                  match.group('tlsa_spec'),
                                                  match.group('tlsa_port'),
                                                  match.group('tlsa_protocol'),
                                                  match.group('tlsa_domain') ),
                                                match.group('pending'),
                                                match.group('time'),
                                                match.group('hash') ) )

            continue

        # delete line
        #   x.com delete 301 25 tcp x.com unix_time count hash
        match = re.match(r'(?P<domain>{})\s+delete\s+(?P<tlsa_spec>{})\s+(?P<tlsa_port>[0-9]+)\s+(?P<tlsa_protocol>{})\s+(?P<tlsa_domain>{})\s+(?P<time>[0-9]+)\s+(?P<count>[0-9]+)\s+(?P<hash>[a-fA-F0-9]+)'.format(prog.tlsa_domain_regex, prog.tlsa_parameters_regex, prog.tlsa_protocol_regex, prog.tlsa_domain_regex), l)
        if match:

            prog.log.info3("  + line {}: delete line (count: {})".format(
                                            line_pos, match.group('count')))

            prog.data.add_line( prog, Prog.DataDelete(
                                            match.group('domain'),
                                            line_pos,
                                            Prog.Tlsa( match.group('tlsa_spec'),
                                                  match.group('tlsa_port'),
                                                  match.group('tlsa_protocol'),
                                                  match.group('tlsa_domain') ),
                                            match.group('count'),
                                            match.group('time'),
                                            match.group('hash') ) )

            continue

        prog.log.error("line {}: malformed line".format(line_pos))
        retval = Prog.RetVal.exit_failure

    prog.log.info3(prog.data)
    return retval


def check_data(prog):
    """Validate the data (from the datafile) in the internal program state.

    Args:
        prog (State): contains the data to check.

    Returns:
        RetVal: return 'RetVal.exit_failure' if errors found,
            'RetVal.ok' if everything is ok.
    """
    prog.log.info3("+++ checking datafile data")
    retval = Prog.RetVal.ok
    # given the prog.data object, which was set by the 'read_datafile'
    # function, check the data in the file.

    # - for every posthook line in a group, there MUST be a prehook line
    #   with (chain.pem|fullchain.pem) if the param is DANE-TA(2) or
    #   (cert.pem|fullchain.pem) if the param is DANE-EE(3).
    # - if there is a posthook line, every prehook line in the group must have
    #   pending set to '1'.
    # - if there are no posthook lines, every prehook line in the group must
    #   have pending set to '0'.
    for g in prog.data.groups:
        if g.post:
            file_list = []
            for l in g.pre:
                file_list += [ l.cert.live.name ]
                if not l.pending == '1':
                    prog.log.error("line {}: pending state on prehook line for domain '{}' conflicts with posthook entry at line {}".format(l.lineno, l.domain, g.post[0].lineno))
                    retval = Prog.RetVal.exit_failure
            for l in g.post:
                if l.tlsa.usage == '2':
                    if not ('chain.pem' in file_list or
                                                'fullchain.pem' in file_list):
                        prog.log.error("line {}: posthook line with tlsa usage '2' has no prehook line with an appropriate certificate to use".format(l.lineno))
                        retval = Prog.RetVal.exit_failure
                else:
                    if not ('cert.pem' in file_list or
                                                'fullchain.pem' in file_list):
                        prog.log.error("line {}: posthook line with tlsa usage '3' has no prehook line with an appropriate certificate to use".format(l.lineno))
                        retval = Prog.RetVal.exit_failure
        else:
            for l in g.pre:
                if not l.pending == '0':
                    prog.log.error("line {}: orphaned prehook line: domain '{}' has no posthook (tlsa) line".format(l.lineno, l.domain))
                    retval = Prog.RetVal.exit_failure

    if retval == Prog.RetVal.ok:
        prog.log.info3("  + datafile ok")

    return retval


def write_prehook(prog):
    """Write to datafile based on prehook mode operation ('prehook lines').

    In prehook mode, the datafile needs to write lines (called
    'prehook lines') that tell subsequent calls of the program what
    certificate live/archive files are, and where. This function will also
    set the permissions of the datafile properly: permissions are changed
    before any meaningful data is written.

    Args:
        prog (State): contains the data to write.

    Returns:
        RetVal: returns 'RetVal.exit_failure' if any errors encountered,
            'RetVal.ok' for success.
    """
    prog.log.info1(
            "+++ writing datafile (prehook): '{}'".format(prog.datafile))

    try:
        prog.datafile.parent.mkdir(parents=True)
    except FileExistsError as ex:
        # Note: this catch can be removed for python 3.5+ since mkdir()
        # accepts the 'exist_ok' parameter.
        if prog.datafile.parent.is_dir():
            # if directory exists, that is fine
            pass
        else:
            prog.log.error(
                "creating datafile directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
            return Prog.RetVal.exit_failure
    except OSError as ex:
        prog.log.error(
            "creating datafile directory '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure


    data = ""

    for t in prog.target_list:
        for c in t.certs:
            prog.log.info3("  + {}\n{}".format(t.domain, c))
            data += '{} "{}" "{}" "{}" 0\n'.format(
                                        t.domain, c.dane, c.live, c.archive)

    if not data:
        prog.log.info1("  + no dane symlinks changed: nothing to write")
        return Prog.RetVal.ok

    header = "# {0} {1}\n# prehook mode {2}, {2:%s}\n".format(
                                        prog.name, prog.version, prog.timenow)


    try:
        with open(str(prog.datafile), "a") as file:
            file.write(header)
    except OSError as ex:
        prog.log.error("writing datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure

    if fix_permissions(prog):
        return Prog.RetVal.exit_failure

    try:
        with open(str(prog.datafile), "a") as file:
            file.write(data)
    except OSError as ex:
        prog.log.error("writing datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure

    return Prog.RetVal.ok


def write_posthook(prog):
    """Write to datafile based on posthook mode operation ('posthook lines').

    In posthook mode, the datafile needs to write lines (called
    'posthook lines') that record what DANE records were published, or need
    to be published or deleted, in addition to appropriate prehook lines,
    which will have been previously read into the internal program state
    object. This function will also set the permissions of the datafile
    properly: permissions are changed before any meaningful data is written.

    Args:
        prog (State): contains the data to write.

    Returns:
        RetVal: returns 'RetVal.exit_failure' if any errors encountered,
            'RetVal.ok' for success.
    """
    prog.log.info1("+++ writing to datafile '{}'".format(prog.datafile))
    data = ""

    for group in prog.data.groups:
        for l in group.pre:
            prog.log.info3(" ++ writing prehook datafile lines...")
            prog.log.info3("{}".format(l))
            if l.state == Prog.DataLineState.write:
                data += '{} "{}" "{}" "{}" {}\n'.format(
                        l.domain, l.cert.dane, l.cert.live, l.cert.archive,
                        l.pending)
        for l in group.post:
            prog.log.info3(" ++ writing posthook datafile lines...")
            prog.log.info3("{}".format(l))
            if l.state == Prog.DataLineState.write:
                data += "{} {}{}{} {} {} {} {} {} {}\n".format(
                        l.domain, l.tlsa.usage, l.tlsa.selector,
                        l.tlsa.matching, l.tlsa.port, l.tlsa.protocol,
                        l.tlsa.domain, l.time, l.pending, l.hash)
        for l in group.special:
            prog.log.info3(" ++ writing delete datafile lines...")
            prog.log.info3("{}".format(l))
            if l.state == Prog.DataLineState.write:
                data += "{} delete {}{}{} {} {} {} {} {} {}\n".format(
                        l.domain, l.tlsa.usage, l.tlsa.selector,
                        l.tlsa.matching, l.tlsa.port, l.tlsa.protocol,
                        l.tlsa.domain, l.time, l.count, l.hash)

    if not data:
        prog.log.info1("  + no data to write")
        return remove(prog)

    header = "# {0} {1}\n# posthook mode {2}, {2:%s}\n".format(
                                        prog.name, prog.version, prog.timenow)

    try:
        with open(str(prog.datafile), "w") as file:
            file.write(header)
    except OSError as ex:
        prog.log.error("writing datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure

    if fix_permissions(prog):
        return Prog.RetVal.exit_failure

    try:
        with open(str(prog.datafile), "a") as file:
            file.write(data)
    except OSError as ex:
        prog.log.error("writing datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure

    return Prog.RetVal.ok


def remove(prog):
    """Remove the datafile, if it exists.

    If the datafile does not exist, we won't look a gift horse in the
    mouth...

    Args:
        prog (State): not modified.

    Returns:
        RetVal: returns 'RetVal.exit_failure' if any errors encountered,
            'RetVal.ok' for success. 
    """
    prog.log.info1("+++ removing datafile '{}'".format(prog.datafile))
    try:
        prog.datafile.unlink()
    except FileNotFoundError as ex:
        pass
    except OSError as ex:
        prog.log.error("removing datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return Prog.RetVal.exit_failure

    return Prog.RetVal.ok


def fix_permissions(prog):
    """Ensure the datafile has the correct permissions.

    'Correct' means mode 0600 and owned by root:root.

    Args:
        prog (State): not modified (except for logging).

    Returns:
        RetVal: returns 'True' if any errors encountered, 'False' for
            success. 
    """
    prog.log.info3(" ++ checking/fixing mode of datafile: should be '0600'")
    try:
        prog.datafile.chmod(0o600)
    except OSError as ex:
        prog.log.error(
            "changing permissions of datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return True

    prog.log.info3(
            " ++ checking/fixing owner of datafile: should be 'root:root'")
    try:
        if not prog.testing_mode:
            os.chown(str(prog.datafile), 0, 0)
    except OSError as ex:
        prog.log.error(
                "changing owner of datafile '{}' failed: {}".format(
                                            ex.filename, ex.strerror.lower()))
        return True

    return False





from setuptools import setup, find_packages

import alnitak

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='alnitak',
    version=alnitak.__version__,
    packages=find_packages(),

    setup_requires=[
        'pytest-runner',
        'requests>=2.21.0',
        'cryptography>=2.4.2',
    ],

    install_requires=[
        'requests>=2.21.0',
        'cryptography>=2.4.2',
    ],

    tests_require=[ 'pytest' ],


    entry_points={ 'console_scripts': [ 'alnitak = alnitak.main:main' ], },

    author='K. S. Kooner',
    author_email='ksa.kooner@gmail.com',
    license='MIT',
    url='https://github.com/definitelyprobably/alnitak',
    description='Create and manage DANE (DNS TLSA) records',
    long_description=long_description,
    keywords="dane tlsa",
    platforms="Linux, POSIX",
    classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "License :: OSI Approved :: MIT License",
            "Operating System :: POSIX :: Linux",
            "Intended Audience :: System Administrators",
            "Topic :: Internet",
            "Topic :: Security",
            "Topic :: System :: Systems Administration",
            "Topic :: Utilities",
        ],

    )




from alnitak.tests import setup
from alnitak import prog

from pathlib import Path
from subprocess import Popen, PIPE



#    Logging schenarios:
#    1.  flags: <NONE>        info   -> logfile
#                             errors -> logfile, stderr
#    2.  flags: -l-           info   -> stdout
#                             errors -> stderr
#    3.  flags: -lno          info   ->X
#                             errors -> stderr
#    4.  flags: -q            info   -> logfile
#                             errors -> logfile
#    5.  flags: -l- -q        info   ->X
#    6.  flags: -lno -q       errors ->X
#
#    7.  flags: -Lno          info   ->X
#                             errors -> logfile, stderr
#    8.  flags: -l- -Lno      info   ->X
#                             errors -> stderr
#    9.  flags: -lno -Lno     info   ->X
#                             errors -> stderr
#    10. flags: -q -Lno       info   ->X
#                             errors -> logfile
#    11. flags: -l- -q -Lno   info   ->X
#    12. flags: -lno -q -Lno  errors ->X



log_lines_both = [
        'error: config file: no targets given',
        '+++ exiting with code: {}'.format(prog.RetVal.config_failure.value),
        ]

log_lines_error = [ 'error: config file: no targets given' ]


def check_for_both(file):
    with open(str(file), 'r') as f:
        lines = f.read().splitlines()
    assert lines[-2:] == log_lines_both

def check_for_error_only(file):
    with open(str(file), 'r') as f:
        lines = f.read().splitlines()
    assert lines[-1:] == log_lines_error


def test_logging1():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l', str(s.varlog / 'log'), '-Lfull',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert log.exists()
    check_for_both(log)

    assert len(stdout) == 0
    assert len(stderr) > 0


def test_logging2():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l-', '-Lfull',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) > 0
    assert len(stderr) > 0


def test_logging3():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-lno', '-Lfull',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) > 0


def test_logging4():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l', str(s.varlog / 'log'), '-Lfull', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert log.exists()
    check_for_both(log)

    assert len(stdout) == 0
    assert len(stderr) == 0


def test_logging5():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l-', '-Lfull', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) == 0


def test_logging6():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-lno', '-Lfull', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) == 0


def test_logging7():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l', str(s.varlog / 'log'), '-Lno',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert log.exists()
    check_for_error_only(log)

    assert len(stdout) == 0
    assert len(stderr) > 0


def test_logging8():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l-', '-Lno',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) > 0


def test_logging9():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-lno', '-Lno',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) > 0


def test_logging10():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l', str(s.varlog / 'log'), '-Lno', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert log.exists()
    check_for_error_only(log)

    assert len(stdout) == 0
    assert len(stderr) == 0


def test_logging11():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-l-', '-Lno', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) == 0


def test_logging12():
    s = setup.Init(keep=True)
    log = Path(s.varlog / 'log')

    assert not log.exists()

    p = Popen(['alnitak', '-t', '-lno', '-Lno', '-q',
            '-c', str(s.configX1)], stdout=PIPE, stderr=PIPE)

    stdout, stderr = p.communicate(timeout=300)

    assert p.returncode == prog.RetVal.config_failure.value

    assert not log.exists()

    assert len(stdout) == 0
    assert len(stderr) == 0


